import tempfile
from configparser import ConfigParser
import mock
import pytest

from xwavecal.main import select_data, run, reduce_data, organize_config, make_output_path, import_obj
from xwavecal.main import RuntimeContext
from xwavecal.tests.utils import FakeImage


def test_select_data():
    header_keys = {'type': 'OBSTYPE',
                   'gain': 'GAIN',
                   'read_noise': 'RDNOISE'}
    type_translator = {'LAMPFLAT': 'lampflat',
                       'DOUBLE': 'wavecal'}
    with tempfile.TemporaryDirectory() as temp_directory:
        files = []
        for i in range(5):
            tmpfile = tempfile.NamedTemporaryFile(suffix='.fits', delete=False, dir=temp_directory)
            files.append(tmpfile.name)
        paths = select_data(temp_directory, 'any', ['.fits'], FakeImage, 0, header_keys, type_translator)
        assert len(paths) == 5
        paths = select_data(temp_directory, 'lampflat', ['.fits'], FakeImage, 0, header_keys, type_translator)
        assert len(paths) == 5
        paths = select_data(temp_directory, 'wavecal', ['.fits'], FakeImage, 0, header_keys, type_translator)
        assert len(paths) == 0
    pass


def test_organize_config():
    config = ConfigParser()
    config.read('xwavecal/tests/data/test_config.ini')
    runtime_context, data_class, extension, header_keys, type_keys = organize_config(config)
    assert data_class == 'xwavecal.images.Image'
    assert extension == 1
    assert type_keys['LAMPFLAT'] == 'lampflat'
    assert type_keys['DOUBLE'] == 'wavecal'
    assert header_keys['read_noise'] == 'RDNOISE'
    assert type(runtime_context.final_wavelength_model) is dict
    assert type(runtime_context.ref_id) is int
    assert type(runtime_context.m0_range) is tuple
    assert runtime_context.time_format == '%Y-%m-%dT%H:%M:%S.%f'


@mock.patch('xwavecal.main.parse_args')
def test_run(mock_args):
    with tempfile.TemporaryDirectory() as temp_directory:
        for i in range(5):
            tempfile.NamedTemporaryFile(suffix='.wrong', delete=False, dir=temp_directory)
        mock_args.return_value = type('', (), {'input_dir': temp_directory, 'output_dir': temp_directory,
                                               'config_file': 'xwavecal/tests/data/test_config.ini',
                                               'frame_type': 'any'})
        run()
    assert True


@mock.patch('xwavecal.main.order_data')
@mock.patch('xwavecal.main.add_data_to_db', return_value=None)
@mock.patch('xwavecal.main.format_db_info', return_value=None)
@mock.patch('xwavecal.main.parse_args')
def test_reduce_data_does_not_err(mock_args, mock_format, mock_add, mock_order):
    config = ConfigParser()
    config.read('xwavecal/tests/data/test_config.ini')
    config.set('stages', 'lampflat', '[]')
    config.set('data', 'data_class', 'xwavecal.tests.utils.FakeImage')
    with tempfile.TemporaryDirectory() as temp_directory:
        tmp = tempfile.NamedTemporaryFile(suffix='.wrong', delete=False, dir=temp_directory)
        mock_order.return_value = [tmp]
        mock_args.return_value = type('', (), {'data_paths': [tmp], 'output_dir': temp_directory,
                                               'frame_type': 'any', 'fpack': False})
        reduce_data(config=config)
    assert True


@mock.patch('xwavecal.main.order_data', return_value=[])
@mock.patch('xwavecal.main.organize_config', return_value=(None, 'xwavecal.images.Image', None, {}, {}))
@mock.patch('configparser.ConfigParser.read')
@mock.patch('xwavecal.main.parse_args')
def test_reduce_data_calls_config(mock_args, mock_config, mock_organize, mock_order):
    mock_args.return_value = type('', (), {'config_file': 'file', 'data_paths': 'path'})
    reduce_data()
    mock_config.assert_called_with('file')
    assert True


def test_make_output_path():
    data = FakeImage()
    path = make_output_path('outdir', data)
    data.set_header_val('site_name', '(test, )')
    path2 = make_output_path('outdir', data)
    assert path == 'outdir/test_nres03_20190410_0077_lampflat_011.fits'
    assert path2 == 'outdir/_test_nres03_20190410_0077_lampflat_011.fits'


def test_import_from_string():
    numpyones = import_obj('numpy.ones')
    assert numpyones((5, 5)).shape == (5, 5)


def test_runtime_context_raises_attr_error():
    with pytest.raises(AttributeError):
        RuntimeContext({}).missing_attribute



from datetime import datetime
import numpy as np
from ast import literal_eval
from configparser import ConfigParser
from scipy.interpolate import UnivariateSpline

from xwavecal.utils.fiber_utils import lit_fibers, lit_wavecal_fibers
from xwavecal.utils.trace_utils import SingleTraceFitter


class FakeContext(object):
    # parses the test_config.ini (which is just the nres_config.ini, but a protected copy for tests only).
    def __init__(self):
        config = ConfigParser()
        config.read('xwavecal/tests/data/test_config.ini')
        dictionary = {key: literal_eval(item) for key, item in config.items('reduction')}
        for attribute, value in dictionary.items():
            setattr(self, attribute, value)


class FakeImage(object):
    def __init__(self, nx=102, ny=100, overscan_size=2, ccdsum='2 2', epoch='20180807'):
        self.nx = nx
        self.ny = ny
        self.telescope_id = -1
        self.site = 'lsc'
        self.instrument = 'nres01'
        self.camera = None
        self.ivar = None
        self.ccdsum = ccdsum
        self.epoch = epoch
        self.overscan_size = overscan_size
        self.data = np.ones((ny, nx))
        self.filename = 'test.fits'
        self.filter = 'U'
        self.dateobs = datetime(2018, 8, 7)
        self.header = {'read_noise': 11, 'gain': 1.0, 'OBSTYPE': 'LAMPFLAT',
                       'rdnoise': 11, 'type': 'lampflat',
                       'observation_date': '2019-04-10T12:56:44.466',
                       'instrument': 'nres03', 'site_name': 'test', 'unique_id': 77,
                       'fiber_state': 'none&tung&tung', 'instrument2': 'fa13'}
        self.filepath = 'None'
        self.caltype = ''
        self.bpm = np.zeros((ny, nx-overscan_size), dtype=np.uint8)
        self.request_number = '0000331403'
        self.block_id = '254478983'
        self.molecule_id = '544562351'
        self.exptime = 30.0
        self.obstype = 'TEST'
        self.data_tables = {}
        self.trace = None
        self.fiber0_lit, self.fiber1_lit, self.fiber2_lit = 0, 1, 1
        self.fiber0_wavecal, self.fiber1_wavecal, self.fiber2_wavecal = 0, 1, 1
        self.wavelength_solution = {}
        self.translator = None

    def get_header_val(self, key):
        return self.header[key]

    def set_header_val(self, key, value):
        self.header[key] = value

    def num_lit_fibers(self):
        return len(lit_fibers(self))

    def num_wavecal_fibers(self):
        return len(lit_wavecal_fibers(self))

    @classmethod
    def load(cls, *args):
        return FakeImage()

    def write(self, *args, **kwargs):
        pass


def gaussian(x, A, b, sigma):
    return A * np.exp(-(x - b) ** 2 / (2 * sigma ** 2))


def noisify_image(image):
    """
    :param image: Banzai_nres FakeImage object.
    This adds poisson and read noise to an image with traces already on it, in that order.
    """
    image.data = np.random.poisson(image.data) + np.random.normal(0, image.get_header_val('RDNOISE'), image.data.shape)


def noisify_spectrum(spectrum, rdnoise=10):
    for i in range(len(spectrum)):
        spectrum['flux'][i] = np.random.poisson(spectrum['flux'][i]) + \
                                np.random.normal(0, rdnoise, spectrum['flux'][i].shape)
    return spectrum


def array_with_peaks(x, centroids, amplitudes, stds):
    vectorized_gaussian = np.vectorize(gaussian)
    y = np.zeros_like(x)
    for centroid, amplitude, std in zip(centroids, amplitudes, stds):
        y += vectorized_gaussian(x, amplitude, centroid, std)
    return y


def fill_image_with_traces(image, poly_order_of_traces=4, order_width=1.5, fiber_intensity=1E4, max_num_traces=1000):
    """
    :param image: Banzai image object, where image.data is a 2d array of the image data
    :param poly_order_of_traces: the max order of the polynomial describing the traces. Maximum of 4.
    :param order_width: width of the traces in pixels
    :param fiber_intensity: peak intensity of the orders
    :param max_num_traces: max number of traces to try and fit onto the image
    :return: An image populated with semi-realistic traces which bend like parabolas, but are of degree
    min(4, poly_order_of_traces).
    """
    trace_fitter = SingleTraceFitter(image_data=image.data,
                                     second_order_coefficient_guess=0,
                                     poly_fit_order=4)
    num_fake_traces = min(int((image.data.shape[1] - 60)/20), max_num_traces)
    coefficients = np.zeros((num_fake_traces, 4+1))
    coefficients[:, 0] = np.linspace(30, image.data.shape[0] - 30, num=num_fake_traces)
    if poly_order_of_traces >= 2:
        coefficients[:, 2] = np.linspace(30, 40, num=num_fake_traces)
    if poly_order_of_traces >= 3:
        coefficients[:, 3] = np.linspace(1, 3, num=num_fake_traces)
    if poly_order_of_traces >= 4:
        coefficients[:, 4] = np.linspace(5, 10, num=num_fake_traces)

    trace_centers = trace_fitter._centers_from_coefficients(coefficients)
    trace_overlay = np.zeros_like(image.data).astype(np.float64)
    vectorized_gaussian = np.vectorize(gaussian)
    for x_pixel in range(trace_centers.shape[1]):
        for i in range(num_fake_traces):
            centroid = trace_centers[i, x_pixel]
            low, high = max(0, int(centroid - 5 * order_width)), min(trace_centers.shape[1] - 1,
                                                                     int(centroid + 5 * order_width)) + 1
            evalwindow = np.arange(low, high, 1)
            if len(evalwindow) > 0:
                trace_overlay[low: high, x_pixel] += vectorized_gaussian(evalwindow, 1, centroid, order_width)
    image.data += trace_overlay * fiber_intensity
    second_order_coefficient_guess = np.mean(coefficients[:, 2])
    return image, trace_centers, second_order_coefficient_guess


class SpectrumUtils:
    @staticmethod
    def _restrict(line_lambdas, lambdas):
        return np.logical_and(line_lambdas >= np.min(lambdas), line_lambdas <= np.max(lambdas))

    @staticmethod
    def uniform_min_space(low, high, size, minspace=0.4):
        out = np.sort(np.random.uniform(low, high, 5*size))
        for i in range(5):
            out = out[1:][~np.isclose(out[1:], out[:-1], atol=minspace)]
        np.random.shuffle(out)
        return out[:size]

    def populate_orders(self, line_lambdas, line_flux, wcs):
        measured_lines = {'pixel': [], 'order': [], 'corrected_flux': [], 'true_wavelength': []}
        pixel = np.arange(wcs.min_pixel, wcs.max_pixel + 1)
        for order in np.arange(wcs.min_order, wcs.max_order + 1):
            order_coords = np.ones_like(pixel) * order
            lam_to_x = UnivariateSpline(wcs(pixel, order_coords), pixel, s=0, k=3)
            in_order = self._restrict(line_lambdas, wcs(pixel, order_coords))
            lines_x_pos = lam_to_x(line_lambdas[in_order])
            measured_lines['pixel'].extend(list(lines_x_pos))
            measured_lines['order'].extend(list(np.ones_like(lines_x_pos) * order))
            measured_lines['corrected_flux'].extend(list(line_flux[in_order]))
            measured_lines['true_wavelength'].extend(list(line_lambdas[in_order]))
        return measured_lines

    def generate_measured_lines(self, n_list, n_true, n_garb, wcs, lam_range=(3800, 9000), min_amp=500, max_amp=1E5):
        """
        :param n_list: number of lines in the line list
        :param n_true: number of lines which are in the spectrum but not in the line list
        :param n_garb: number of lines which are not true emission lines. These will
        be injected randomly per order. E.g. cosmic rays.
        :param wcs: wavelength solution to follow to generate the line positions
        :return: Dictionary: measured_lines with n_list + n_true + n_garb lines.
                 measured_lines['pixel'] is the pixel position of the line centroid.
                 measured_lines['order'] is the order index of the line.
                 measured_lines['corrected_flux'] is the flux of the line, where duplicated lines
                 have the same flux.
                 measured_lines['true_wavelength'] is the true wavelength of the line. For garbage
                 lines, this is np.nan.
        """
        # line lambda locations
        full_list = self.uniform_min_space(min(lam_range), max(lam_range), size=n_list + n_true)
        line_list, line_other = full_list[:n_list], full_list[:n_true]
        # fluxes
        line_list_flux = np.random.randint(int(min_amp), int(max_amp), size=n_list)
        line_other_flux = np.random.randint(int(min_amp), int(max_amp), size=n_true)
        measured_lines = self.populate_orders(np.hstack([line_list, line_other]),
                                              np.hstack([line_list_flux, line_other_flux]), wcs)
        # adding garbage lines
        line_garbage_x = np.random.randint(wcs.min_pixel, wcs.max_pixel, size=n_garb).astype(np.float32)
        line_garbage_o = np.random.randint(wcs.min_order, wcs.max_order+1, size=n_garb)
        line_garbage_flux = np.random.randint(int(min_amp), int(max_amp), size=n_garb)
        measured_lines['pixel'].extend(list(line_garbage_x))
        measured_lines['order'].extend(list(line_garbage_o))
        measured_lines['corrected_flux'].extend(list(line_garbage_flux))
        measured_lines['true_wavelength'].extend(list(np.zeros(n_garb, dtype=float) * np.nan))
        # cast the elements of measured lines into arrays
        for key, item in measured_lines.items():
            measured_lines[key] = np.array(item)
        return measured_lines, line_list




import sqlite3
import os
from datetime import datetime

import logging
logger = logging.getLogger(__name__)


def add_data_to_db(db_path, db_info):
    db_exists = os.path.exists(db_path)
    conn = sqlite3.connect(db_path)
    c = conn.cursor()
    if not db_exists:
        logger.info('No existing database found at {0}. Creating a new one at {1}.'.format(db_path, db_path))
        # Create table
        c.execute('''CREATE TABLE caldata (id integer primary key autoincrement, 
                     type text, observation_date text,
                     date_created text, instrument text, instrument2 text, site_name text, 
                     fiber0 integer, fiber1 integer, fiber2 integer, 
                     is_bad integer, filepath text)''')
    match = query_db_for_match(c, db_info)
    if match is None:
        entries = tuple(db_info[key] for key in ['type', 'observation_date', 'date_created', 'instrument',
                                                 'instrument2',
                                                 'site_name', 'fiber0', 'fiber1', 'fiber2', 'is_bad',
                                                 'filepath'])
        c.execute('INSERT INTO caldata (type, observation_date, date_created, instrument, instrument2, site_name,'
                      ' fiber0, fiber1, fiber2, is_bad, filepath) '
                      'VALUES (?,?,?,?,?,?,?,?,?,?,?)', entries)
    else:
        c.execute('UPDATE caldata SET date_created=?, is_bad=?, filepath=? WHERE id=?',
                  (db_info['date_created'], db_info['is_bad'], db_info['filepath'], match))
    conn.commit()
    conn.close()


def format_db_info(data, time_format='%Y-%m-%dT%H:%M:%S.%f'):
    db_info = {'type': data.get_header_val('type'),
               'observation_date': datetime.strptime(data.get_header_val('observation_date'), time_format),
               'date_created': datetime.now(),
               'instrument': data.get_header_val('instrument'),
               'instrument2': data.get_header_val('instrument2'),
               'site_name': data.get_header_val('site_name'),
               'fiber0': int(data.fiber0_lit),
               'fiber1': int(data.fiber1_lit),
               'fiber2': int(data.fiber2_lit),
               'is_bad': 0, 'filepath': data.filepath}
    return db_info


def query_db_for_nearest(db_path, data, type_to_find, time_format='%Y-%m-%dT%H:%M:%S.%f', date=None):
    if not os.path.exists(db_path):
        logger.error('No existing database found at {0}. Cannot fetch calibration.'.format(db_path, db_path))
        return None
    db_info = format_db_info(data, time_format)
    if date is None:
        date = db_info['observation_date']
    conn = sqlite3.connect(db_path)
    c = conn.cursor()
    search_criteria = (type_to_find, db_info['instrument'], db_info['instrument2'], db_info['site_name'],
                       db_info['fiber0'], db_info['fiber1'], db_info['fiber2'],
                       date)
    c.execute('SELECT filepath FROM caldata WHERE is_bad=0 AND type=? AND instrument=? AND instrument2=?'
              'AND site_name=? AND fiber0=? AND fiber1=? AND fiber2=?'
              'ORDER BY ABS( strftime("%s", observation_date) - strftime("%s", ?) ) ASC',
               search_criteria)
    results = c.fetchall()
    conn.close()
    if len(results) == 0:
        return None
    else:
        return results[0][0]


def query_db_for_match(cursor, db_info):
    search_criteria = tuple(db_info[key] for key in ['type', 'observation_date',
                                                     'instrument', 'instrument2', 'site_name',
                                                     'fiber0', 'fiber1', 'fiber2'])
    cursor.execute('SELECT id FROM caldata WHERE type=? AND observation_date=?'
                   'AND instrument=? AND instrument2=? AND site_name=? AND fiber0=? AND fiber1=? AND fiber2=?',
                    search_criteria)
    results = cursor.fetchall()
    if len(results) == 0:
        return None
    else:
        return results[0][0]



"""
Module for rotating and flipping frames

This is to make frames such that the mean wavelengths of the diffraction orders increase
from left to right (e.g bluest horizontal coordinate is x=0 and reddest is x=4096)
and increase from top to bottom (e.g. bluest vertical coordinate is y=4096 and reddest is y=0).
"""

import numpy as np

from xwavecal.stages import Stage


class Rot90(Stage):
    """
    rotate image.data counter-clockwise by 90 degrees.
    """
    def __init__(self, runtime_context):
        super(Rot90, self).__init__(runtime_context)

    def do_stage(self, image):
        image.data = np.rot90(image.data)
        image.ivar = None if image.ivar is None else np.rot90(image.ivar)
        return image


class FlipHoriz(Stage):
    def __init__(self, runtime_context):
        super(FlipHoriz, self).__init__(runtime_context)

    def do_stage(self, image):
        image.data = np.flip(image.data, axis=1)
        image.ivar = None if image.ivar is None else np.flip(image.ivar, axis=1)
        return image


class FlipVert(Stage):
    def __init__(self, runtime_context):
        super(FlipVert, self).__init__(runtime_context)

    def do_stage(self, image):
        image.data = np.flip(image.data, axis=0)
        image.ivar = None if image.ivar is None else np.flip(image.ivar, axis=0)
        return image



import mock
import numpy as np
from astropy.table import Table

import xwavecal.utils.overlap_utils as overlapu
from xwavecal.utils.overlap_utils import OverlapFitter


class Utils:
    @staticmethod
    def fake_overlap(b_lines, r_lines, b_flux, r_flux, linearscale, *args, **kwargs):
        return {'pixel': r_lines, 'matched_pixel': b_lines, 'peaks': 1}


class TestOverlapFitter:
    def test_fit_overlap(self):
        true_coeffs = [100, 1.4, 2E-5]
        num_peaks = 10
        r_lines = np.random.uniform(0, 200, num_peaks)
        r_flux = np.random.randint(100, int(1E5), num_peaks)
        b_lines = overlapu.coordinate_transform(r_lines, true_coeffs)
        b_flux = 1. * r_flux
        coeffs = OverlapFitter()._fit_overlap(b_lines, r_lines, b_flux, r_flux, (0.5, 2), pixel_tol=2, flux_tol=0.2, deg=2)
        assert np.allclose(coeffs, true_coeffs, rtol=0.05)

    @mock.patch('xwavecal.utils.overlap_utils.OverlapFitter._fit_overlap', return_value=[100, 1.4, 2E-5])
    def test_peaks_save(self, mock_fit):
        r_lines = np.random.randint(0, 100, 20)
        b_lines = overlapu.coordinate_transform(r_lines, [100, 1.4, 2E-5])
        overlap = OverlapFitter().fit(b_lines, r_lines, None, None, None)
        assert np.allclose(overlap['matched_pixel'], b_lines)
        assert np.allclose(overlap['pixel'], r_lines)
        assert np.isclose(overlap['peaks'], 20)

    @mock.patch('xwavecal.utils.overlap_utils.OverlapFitter.fit', side_effect=Utils.fake_overlap)
    def test_fit_overlaps(self, mock_overlap):
        spectrum = {'flux': np.zeros((3, 3)), 'pixel': np.array([np.arange(3), np.arange(1, 4), np.arange(3)]),
                    'ref_id': np.arange(1, 4)}
        lines = {'corrected_flux': np.zeros(spectrum['pixel'].size), 'pixel': spectrum['pixel'].flatten(),
                 'order': (np.arange(1, 4).reshape(-1, 1) * np.ones_like(spectrum['pixel'])).flatten()}
        overlaps = overlapu.fit_overlaps(spectrum, lines, max_overlap_red=2, max_overlap_blue=2,
                                         linear_scale_range=(0.5, 2), deg=-1)
        for i in range(2):
            assert np.allclose(overlaps['pixel'][i][0], spectrum['pixel'][i][0])
            assert np.allclose(overlaps['matched_pixel'][i][0], spectrum['pixel'][i+1][-1])
            assert np.all(np.isnan(overlaps['pixel'][i][-1]))
            assert np.all(np.isnan(overlaps['matched_pixel'][i][-1]))
            assert np.isclose(overlaps['peaks'][i], 1)
            assert np.isclose(overlaps['ref_id'][i], i + 1)
            assert np.isclose(overlaps['matched_ref_id'][i], i + 2)

    def test_fit_overlaps_skips_on_no_lines(self):
        spectrum = {'flux': np.zeros((3, 3)), 'pixel': np.array([np.arange(3), np.arange(1, 4), np.arange(3)]),
                    'ref_id': np.arange(1, 4)}
        lines = {'corrected_flux': np.zeros(spectrum['pixel'].size), 'pixel': spectrum['pixel'].flatten(),
                 'order': (np.arange(1, 4).reshape(-1, 1) * np.ones_like(spectrum['pixel'])).flatten()}
        overlaps = overlapu.fit_overlaps(spectrum, lines, max_overlap_red=2, max_overlap_blue=2,
                                         linear_scale_range=(0.5, 2), deg=10)
        assert len(overlaps) == 0

    def test_flag_bad_overlaps(self):
        overlaps = Table({'peaks': np.array([[1], [0], [0], [1]]), 'good': [True, True, True, True]})
        overlaps = overlapu.flag_bad_overlaps(overlaps, 1)
        assert np.allclose(overlaps['good'], [True, False, False, True])



from astropy.io import fits
from astropy.table import Table

from xwavecal.utils.fiber_utils import fiber_states_from_header, wavecal_fibers_from_header
from xwavecal.utils.fiber_utils import lit_wavecal_fibers, lit_fibers
from xwavecal.utils import fits_utils

import logging
logger = logging.getLogger(__name__)


class DataProduct(object):
    """
    Base class with .write and .load methods for all data products.
    """
    def __init__(self, data=None, data_name=None, filepath=None, header=None,
                 translator=None):
        if header is None:
            header = {}
        self.header = header
        self.filepath = filepath
        self.data = Table(data)
        self.data_name = data_name
        self.translator = translator

    def _get_header_vals(self, keys):
        keys = [keys] if isinstance(keys, str) else keys
        out = tuple(self.header[key] for key in keys)
        return out if len(out) > 1 else out[0]

    def get_header_val(self, key):
        if self.translator is not None:
            keys = self.translator[key]
            return self.translator.type_keys.get(self._get_header_vals(keys), self._get_header_vals(keys))
        return self._get_header_vals(key)

    def set_header_val(self, key, value):
        if self.translator is not None:
            key = self.translator[key]
        self.header[key] = value

    def write(self, fpack=False):
        hdu = fits.BinTableHDU(self.data, name=self.data_name, header=fits.Header(self.header))
        hdu_list = fits.HDUList([fits.PrimaryHDU(), hdu])
        self._update_filepath(fpack)
        logger.info('Writing file to {filepath}'.format(filepath=self.filepath))
        fits_utils.writeto(hdu_list=hdu_list, filepath=self.filepath, fpack=fpack,
                           overwrite=True, output_verify='fix+warn')

    def _update_filepath(self, fpack=False):
        if fpack and not self.filepath.endswith('.fz'):
            self.filepath += '.fz'

    @classmethod
    def load(cls, path, extension_name, translator=None):
        hdu_list = fits.open(path)
        return cls(data=hdu_list[extension_name].data, header=hdu_list[extension_name].header,
                   filepath=path, translator=translator)


class Image(DataProduct):
    """
    Class for two-dimensional data frames.
    """
    def __init__(self, filepath=None, data=None, header=None, data_tables=None, translator=None, trace=None,
                 data_name=None, ivar=None, wavelength_solution=None):
        super(Image, self).__init__(filepath=filepath, data=data, header=header, translator=translator, data_name=data_name)
        if data_tables is None:
            data_tables = {}
        if wavelength_solution is None:
            wavelength_solution = {}
        self.data_tables = data_tables
        self.data = data
        self.ivar = ivar
        self.filepath = filepath
        self.trace = trace
        self.rectified_2d_spectrum = None
        self.rectified_ivar = None
        self.fiber0_lit, self.fiber1_lit, self.fiber2_lit = fiber_states_from_header(self.get_header_val('fiber_state'))
        self.fiber0_wavecal, self.fiber1_wavecal, self.fiber2_wavecal = wavecal_fibers_from_header(self.get_header_val('fiber_state'))
        self.wavelength_solution = wavelength_solution

    def write(self, fpack=False):
        if self.data_name is None:
            hdus = [fits.PrimaryHDU(data=self.data, header=fits.Header(self.header))]
        else:
            hdus = [fits.PrimaryHDU(), fits.ImageHDU(data=self.data, header=fits.Header(self.header), name=self.data_name)]
        if self.ivar is not None:
            hdus.append(fits.ImageHDU(data=self.ivar, header=None, name='ivar'))

        table_hdus = [fits.BinTableHDU(table, name=name, header=table.meta.get('header'))
                      for name, table in self.data_tables.items()]
        hdu_list = fits.HDUList([*hdus, *table_hdus])
        self._update_filepath(fpack)
        logger.info('Writing file to {filepath}'.format(filepath=self.filepath))
        fits_utils.writeto(hdu_list=hdu_list, filepath=self.filepath, fpack=fpack,
                           overwrite=True, output_verify='fix+warn')

    def num_lit_fibers(self):
        return len(lit_fibers(self))

    def num_wavecal_fibers(self):
        return len(lit_wavecal_fibers(self))

    @classmethod
    def load(cls, path, extension_name, translator=None):
        hdu_list = fits.open(path)
        data_tables = {hdu.name: Table(hdu.data) for hdu in hdu_list if type(hdu) is fits.BinTableHDU}
        return cls(data=hdu_list[extension_name].data, header=hdu_list[extension_name].header,
                   filepath=path, translator=translator, data_tables=data_tables)


class SplitHeaderImage(Image):
    """
    Functionally the same as xwavecal.images.Image , except this assumes that the [0] extension
    of the raw data .fits file has information in the header integral to reduction. The [0].header is
    appended onto the primary_extension (e.g. [1], specified in the config.ini) header for the image.
    Example: This class is suitable for HARPS data.
    """
    def __init__(self, filepath=None, data=None, header=None, data_tables=None, translator=None, trace=None,
                 data_name=None, ivar=None):
        super(SplitHeaderImage, self).__init__(filepath=filepath, data=data, header=header, translator=translator,
                                               data_name=data_name, data_tables=data_tables, trace=trace, ivar=ivar)

    @classmethod
    def load(cls, path, extension_name, translator=None):
        hdu_list = fits.open(path)
        data_tables = {hdu.name: Table(hdu.data) for hdu in hdu_list if type(hdu) is fits.BinTableHDU}
        header = hdu_list[extension_name].header
        header.extend(hdu_list[0].header)
        return cls(data=hdu_list[extension_name].data, header=header, filepath=path, translator=translator,
                   data_tables=data_tables)


import mock
import numpy as np

import tempfile
import os
from astropy.table import Table
from astropy.io.fits import Header


from xwavecal.images import Image, DataProduct
from xwavecal.utils.fits_utils import Translator
from xwavecal.tests.utils import FakeContext, FakeImage


def test_image_class_loads():
    image = Image(header={'fiber_state': 'tung&tung&none'})
    assert image.trace is None
    assert image.rectified_2d_spectrum is None
    assert image.wavelength_solution == {}


def test_get_num_lit_fibers():
    image = Image(header={'fiber_state': 'tung&tung&none'})
    assert image.num_lit_fibers() == 2
    image = Image(header={'fiber_state': 'none&tung&none'})
    assert image.num_lit_fibers() == 1


def test_get_num_lit_wavecal_fibers():
    image = Image(header={'fiber_state': 'thar&tung&none'})
    assert image.num_wavecal_fibers() == 1
    assert image.fiber0_wavecal == 1
    assert image.fiber1_wavecal == 0
    image = Image(header={'fiber_state': 'none&thar&thar'})
    assert image.num_wavecal_fibers() == 2
    assert image.fiber1_wavecal == 1
    assert image.fiber2_wavecal == 1


class TestDataProduct:
    def test_load_and_write(self):
        name = 'trace'
        image = DataProduct(data={'id': [1], 'centers': [np.arange(3)]}, data_name='trace')
        with tempfile.TemporaryDirectory() as tmp_directory_name:
            path = os.path.join(tmp_directory_name, 'test_trace_table.fits')
            image.filepath = path
            image.header = {'bla': 1}
            image.write(False)
            loaded_image = DataProduct.load(path=path, extension_name=name)
            assert np.allclose(loaded_image.data['centers'][0], image.data['centers'][0])
            assert np.allclose(loaded_image.data['id'][0], image.data['id'][0])
            assert np.isclose(loaded_image.get_header_val('bla'), 1)

    def test_write_gets_correct_filename(self):
        name = 'trace'
        image = DataProduct(data=Table({'id': [1], 'centers': [np.arange(3)]}), data_name=name)
        with tempfile.TemporaryDirectory() as tmp_directory_name:
            for fpack, extension in zip([True, False], ['.fz', 'its']):
                path = os.path.join(tmp_directory_name, 'test_trace_table.fits')
                image.filepath = path
                image.header = {'bla': 1}
                image._update_filepath(fpack)
                assert image.filepath[-3:] == extension

    def test_translator(self):
        header_keys = {'type': 'OBSTYPE',
                       'gain': 'GAIN',
                       'read_noise': 'RDNOISE',
                       'multi': ('obstype', 'gain')}
        type_keys = {'LAMPFLAT': 'lampflat',
                     'DOUBLE': 'wavecal'}
        translator = Translator(header_keys, type_keys)
        header = Header({'OBSTYPE': 'DOUBLE', 'GAIN': 1, 'RDNOISE': 10})
        image = DataProduct(header=header, translator=translator)
        assert image.get_header_val('type') == 'wavecal'
        assert image.get_header_val('gain') == 1
        assert image.get_header_val('read_noise') == 10
        assert image.get_header_val('multi') == ('DOUBLE', 1)
        image.set_header_val('gain', 5)
        assert image.get_header_val('gain') == 5


class TestImage:
    def test_load_and_write(self):
        image = Image(data=np.ones((10, 10)), header={'bla': 1, 'fiber_state': 'none&none&none'})
        with tempfile.TemporaryDirectory() as tmp_directory_name:
            path = os.path.join(tmp_directory_name, 'test_trace_table.fits')
            image.filepath = path
            image.write(fpack=False)
            image = Image.load(path=path, extension_name=0)
            assert np.allclose(image.data, 1)
            assert image.header['bla'] == 1

    def test_load_and_write_with_name(self):
        name = 'a_name'
        image = Image(data=np.ones((10, 10)), data_name='a_name', header={'bla': 1, 'fiber_state': 'none&none&none'})
        with tempfile.TemporaryDirectory() as tmp_directory_name:
            path = os.path.join(tmp_directory_name, 'test_trace_table.fits')
            image.filepath = path
            image.write(fpack=False)
            image = Image.load(path=path, extension_name=name)
            assert np.allclose(image.data, 1)
            assert image.header['bla'] == 1



"""
blaze.py: Module for making blaze calibration files for xwavecal spectrographs.
"""
import numpy as np
from copy import deepcopy

from xwavecal.stages import Stage, ApplyCalibration
from xwavecal.images import Image
from xwavecal.utils.blaze_utils import normalize_orders


class ApplyBlaze(ApplyCalibration):
    """
    Divides input images by the per-order normalized blaze spectrum.
    """
    def __init__(self, runtime_context):
        super(ApplyBlaze, self).__init__(runtime_context)

    @property
    def calibration_type(self):
        return 'BLAZE'

    def apply_master_calibration(self, image, master_calibration_path):
        blaze = Image.load(master_calibration_path, extension_name=self.runtime_context.blaze_name,
                           translator=image.translator)
        if not np.allclose(image.data.shape, blaze.data.shape):
            self.logger.error('Shape of blaze data and image data do not agree. Aborting blaze correction. Wavelength'
                         'solution may suffer.')  # pragma: no cover
        else:
            self.logger.info('Dividing by blaze from file')
            image.ivar = None if image.ivar is None else image.ivar * np.power(blaze.data, 2)
            # TODO full error propagation. The above does not hold for low (< 5) signal-to-noise pixels.
            image.data = np.divide(image.data, blaze.data, out=image.data.astype(float), where=~np.isclose(blaze.data, 0))
            image.set_header_val('IDBLAZE', (master_calibration_path, 'ID of blaze file used'))
        return image


class BlazeMaker(Stage):
    def __init__(self, runtime_context):
        super(BlazeMaker, self).__init__(runtime_context)

    @property
    def calibration_type(self):
        return 'BLAZE'

    def do_stage(self, image):
        self.logger.info('Making blaze file.')
        blaze = Image(data=deepcopy(image.data), translator=image.translator,
                      trace=image.trace, header=deepcopy(image.header), data_name=self.runtime_context.blaze_name)
        blaze.data = blaze.data / normalize_orders(blaze.data, blaze.trace, half_window=self.runtime_context.max_extraction_half_window)
        if image.ivar is not None:
            blaze.data[image.data * np.sqrt(image.ivar) < self.runtime_context.min_blaze_sn] = 0
        blaze.set_header_val('type', self.calibration_type.lower())
        return image, blaze



import numpy as np
from copy import deepcopy

from xwavecal.utils.misc_utils import normalize_by_brightest, n_largest_elements
from xwavecal.utils.correlate import correlate2d
from xwavecal.utils.fiber_utils import lit_wavecal_fibers, lit_fibers
from xwavecal.stages import ApplyCalibration, Stage
from xwavecal.images import DataProduct

import logging
logger = logging.getLogger(__name__)


class MakeFiberTemplate(Stage):
    def __init__(self, runtime_context=None):
        super(MakeFiberTemplate, self).__init__(runtime_context=runtime_context)

    @property
    def calibration_type(self):
        return 'FIBERS'

    def do_stage(self, image):
        order = self.runtime_context.template_trace_id
        self.logger.info('Generating a template with center order {0}'.format(order))
        spec = image.data_tables[self.runtime_context.main_spectrum_name]
        num_fibers = len(lit_fibers(image))
        orders = [order - num_fibers, order, order + num_fibers]
        template_region = np.logical_or.reduce(tuple(spec['id'] == i for i in orders))
        template = DataProduct(data=spec[template_region], header=deepcopy(image.header),
                               data_name=self.calibration_type.lower(), translator=image.translator)
        template.set_header_val('type', self.calibration_type.lower())
        template.fiber0_lit, template.fiber1_lit, template.fiber2_lit = image.fiber0_lit, image.fiber1_lit, image.fiber2_lit
        return image, template


class IdentifyFibers(ApplyCalibration):
    def __init__(self, runtime_context=None):
        super(IdentifyFibers, self).__init__(runtime_context=runtime_context)

    @property
    def calibration_type(self):
        return 'FIBERS'

    def apply_master_calibration(self, image, template_path):
        """
        :param image: DataProduct
        :param template_path: path to the arc lamp template which will be used to identify the fibers
        :return: image. The spectrum data_table has two columns appended to it: fiber and ref_id.

        NOTE: We follow the convention that the lowest lying ThAr fiber (i.e. has smallest x and y coordinate
        on the detector) corresponds to the smallest lit fiber index. E.g. if 0, 1 are lit with ThAr, then the lowest
        lying ThAr fiber is denoted 0. If 1,2 are lit, then the lowest lying ThAr fiber is denoted 1.
        ref_ids must be such that the mean wavelength of an order always increases with increasing ref_id

        NOTE: This stage relies on the fact that only 2 fibers are lit at one time.
        """
        if len(image.data_tables[self.runtime_context.main_spectrum_name]['flux']) == 0:
            logger.error('Image has length 0 spectrum. Skipping fiber identification')
        elif image.num_wavecal_fibers() == 0:
            logger.info('Image does not have any fibers lit with ThAr, skipping fiber identification.', )
        else:
            logger.info('Identifying fibers via cross correlation.')
            spectrum = image.data_tables[self.runtime_context.main_spectrum_name]

            read_noise = image.get_header_val('read_noise')
            signal_to_noise = spectrum['flux'] / (np.sqrt(np.abs(spectrum['flux']) + read_noise ** 2))
            spectrum_to_search = normalize_by_brightest(signal_to_noise)

            template = self.construct_single_fiber_template(template_path,
                                                            self.calibration_type.lower(),
                                                            num_lit_fibers=image.num_lit_fibers())
            matched_ids = self.identify_matching_orders(spectrum_to_search, template,
                                                        num_arc_lamp_fibers=image.num_wavecal_fibers())

            fiber_ids = self.build_fiber_column(matched_ids, lit_fibers(image), lit_wavecal_fibers(image), len(spectrum['id']),
                                                low_fiber_first=getattr(self.runtime_context, 'low_fiber_first', True))
            ref_ids = self.build_ref_id_column(matched_ids, fiber_ids, self.runtime_context.ref_id,
                                               low_fiber_first=getattr(self.runtime_context, 'low_fiber_first', True))
            # TODO refactor storing information about which spectra exist.
            image.set_header_val('IDTEMPL', (template_path, 'ID of the fiber template.'))
            for key in [self.runtime_context.main_spectrum_name, self.runtime_context.blaze_corrected_spectrum_name]:
                if image.data_tables.get(key) is not None:
                    image.data_tables[key]['fiber'] = fiber_ids
                    image.data_tables[key]['ref_id'] = ref_ids
        return image

    @staticmethod
    def build_fiber_column(matched_ids, fibers, wavecal_fibers, num_traces, low_fiber_first=True):
        """
        :param matched_ids: array: the indices of spectrum who are spectrally matched to ref_id
        :param fibers: ndarray. E.g. [0, 1] if fibers 0 and 1 are lit. [0,1,2,3] if fibers 0 through 3 are lit etc..
        :param wavecal_fibers: ndarray. E.g. [0, 1] if fibers 0 and 1 are lit with a wavelength calibration source.
        :param num_traces: int. Number of lit traces, i.e. number of rows in the extracted spectrum.
        :param low_fiber_first: whether the lowest lying fiber (lowest in terms of row index in spectrum)
        should be associated with the first lit fiber. For example:
           if lit_wavecal_fibers(image) = [1, 2] (fibers 1 and 2 are lit with a wavelength calibration lamp)
           and matched_ids = [10, 11] (the rows 10 and 11 in the spectrum have been matched to the template)
           then if label_low_fiber:
              spectrum[10]['fiber_id'] will equal 1, and spectrum[11]['fiber_id'] will equal 2
           If not label_low_fiber:
              spectrum[10]['fiber_id'] will equal 2, and spectrum[11]['fiber_id'] will equal 1

        :return: array:
                 array of fiber identification numbers such that spectrum[fiber_ids == 1] would yield a complete
                 extracted spectrum for the 1 fiber on the echelle spectrograph. spectrum[fiber_ids == 2] would
                 be the spectrum for fiber 2.
        """
        fiber_sequence = fibers if low_fiber_first else fibers[::-1]
        match_fiber = wavecal_fibers[0] if low_fiber_first else wavecal_fibers[-1]
        num_lit = len(fibers)
        sequence_ids = np.arange(num_traces) % num_lit
        sequence_ids += np.where(fiber_sequence == match_fiber)[0] - sequence_ids[np.min(matched_ids)]
        return fiber_sequence[sequence_ids % num_lit]

    @staticmethod
    def build_ref_id_column(matched_ids, fiber_ids, ref_id, low_fiber_first=True):
        """
        :param matched_ids: array: the indices of spectrum who are spectrally matched to ref_id
        :param ref_id: the reference id which is to be assigned to those orders which have been spectrally
        matched with the template and whose row id is given in matched_ids.
        :param fiber_ids: array: integers which designate to each row of spectrum, a fiber id. See
        IdentifyFibers.build_fiber_column().
        :param low_fiber_first: bool. See IdentifyFibers.build_fiber_column()
        :return: ref_ids. array.
                 array of reference ids (order indices), each entry corresponding
                 to each row of the spectrum, where ref_ids[j] is the i in the 1/(m0+i) of the grating equation,
                 to be used with the jth diffraction order (indexed from 0). E.g. order 0 has i = ref_ids[0].
                 We assume that spectrum is such that spectrum[i] is bluer than spectrum[j] if i > j.
                 Thus, ref_ids[i] > ref_ids[j] (where ref_ids[i] is the reference id of the single order spectrum[i]).
                 This array is such that ref_ids[matched_ids[k]] = ref_id for all valid indices k
                 of matched_ids.
        """
        fiber_sequence = list(set(fiber_ids)) if low_fiber_first else list(set(fiber_ids))[::-1]
        group_start = np.where(fiber_ids == fiber_sequence[0])[0][0]
        ref_ids = np.zeros_like(fiber_ids)
        num_lit = len(set(fiber_ids))
        for i in range(num_lit):
            ref_ids[i + group_start::num_lit] = np.arange(1, len(ref_ids[i + group_start::num_lit]) + 1)
        ref_ids += ref_id - ref_ids[np.min(matched_ids)]
        return ref_ids

    @staticmethod
    def construct_single_fiber_template(template_path, ext_name, num_lit_fibers=2):
        template = DataProduct.load(template_path, extension_name=ext_name)
        single_fiber = template.data['flux'].data
        template = np.zeros((num_lit_fibers * (single_fiber.shape[0] - 1) + 1, single_fiber.shape[1]))
        template[::num_lit_fibers] = single_fiber
        return template

    @staticmethod
    def identify_matching_orders(two_d_spectrum, template, num_arc_lamp_fibers=2):
        order_likelyhood = correlate2d(two_d_spectrum, template, max_lag=100)
        one_d_likelyhood = np.max(order_likelyhood, axis=1)
        matched_ids = n_largest_elements(one_d_likelyhood, n=num_arc_lamp_fibers)
        return np.array(matched_ids, dtype=np.int)



from configparser import ConfigParser
import tempfile
import os
import mock
import numpy as np
from glob import glob
import pytest
from astropy.io import fits
from astropy.table import Table

from xwavecal.main import reduce_data, organize_config

@pytest.mark.e2e
@mock.patch('xwavecal.fibers.IdentifyFibers.get_calibration_filename',
            return_value='xwavecal/tests/data/nres_test_data/cpt_nres03_20190405_0014_fibers_011.fits')
def test_reduce_data(mock_arc_template):
    # TODO this test should be replaced with a proper pytest fixture which makes the tempfile
    #  once per session. Then we can make independent tests for lampflat and arc creation.
    with tempfile.TemporaryDirectory() as temp_directory:
        args = type('test_args', (), {'input_dir': 'xwavecal/tests/data/nres_test_data/',
                                      'output_dir': temp_directory, 'fpack': False})
        config = ConfigParser()
        config.read('xwavecal/tests/data/test_config.ini')
        config.set('reduction', 'database_path', '"' + os.path.join(temp_directory, 'test.db') + '"')
        data_paths = glob('xwavecal/tests/data/nres_test_data/*w00*.fits*')
        reduce_data(data_paths, args=args, config=config)
        # check the reduced trace, lampflat and blaze.
        # fetch the context for easy access to configuration information
        context = organize_config(config)[0]
        # check that the lampflat, traces and blaze exist
        trace_file = glob(os.path.join(temp_directory, '*trace*'))[0]
        assert os.path.exists(glob(os.path.join(temp_directory, '*lampflat*'))[0])
        assert os.path.exists(trace_file)
        assert os.path.exists(glob(os.path.join(temp_directory, '*blaze*'))[0])
        check_traces(trace_file, context)

        # reduce the arc lamp file and make a wavelength solution.
        data_paths = glob('xwavecal/tests/data/nres_test_data/*a00*.fits*')
        reduce_data(data_paths, args=args, config=config)
        # check the wavelength calibration
        wavecal_file = glob(os.path.join(temp_directory, '*wavecal*'))[0]
        assert os.path.exists(wavecal_file)
        check_wavelength_calibration(wavecal_file, context)


def check_wavelength_calibration(fname, context):
    with fits.open(fname) as wcs:
        # assert that both fibers have at least 1000 lines with residuals less than 0.004 Angstrom.
        lines = Table(wcs[context.emission_lines_table_name].data)
        both_fibers = [np.isclose(lines['fiber'], 1), np.isclose(lines['fiber'], 2)]
        for fiber in both_fibers:
            assert 1000 < np.count_nonzero(np.abs(lines[fiber]['wavelength'] - lines[fiber]['reference_wavelength']) < 0.004)
        # assert that both fibers have >25 overlaps with peaks>6 and >15 marked as good.
        overlaps = Table(wcs[context.overlap_table_name].data)
        both_fibers = [np.isclose(overlaps['fiber'], 1), np.isclose(overlaps['fiber'], 2)]
        for fiber in both_fibers:
            assert len(overlaps[fiber]['peaks'] >= 6) > 25
            assert len(overlaps[fiber]['good'] >= 6) > 15


def check_traces(fname, context):
    with fits.open(fname) as trace:
        assert len(Table(trace[context.trace_table_name].data)) == 135


import numpy as np

from xwavecal.tests.utils import FakeImage, FakeContext
from xwavecal.variance import CalcInverseVariance


class TestCalcInverseVariance:
    def test_calc_zero_rdnoise(self):
        image = FakeImage()
        image.header = {'read_noise': 0}
        image.data = np.random.random((10, 10)) * 100
        image = CalcInverseVariance(FakeContext()).do_stage(image)
        assert np.allclose(image.ivar, image.data ** (-1))

    def test_calc_zero_signal(self):
        image = FakeImage()
        image.header = {'read_noise': 10}
        image.data = np.zeros((10, 10))
        image = CalcInverseVariance(FakeContext()).do_stage(image)
        assert np.allclose(image.ivar, image.get_header_val('read_noise') ** (-2))

    def test_calc(self):
        image = FakeImage()
        image.header = {'read_noise': 10}
        image.data = np.random.random((10, 10)) * 100 - 50
        image = CalcInverseVariance(FakeContext()).do_stage(image)
        assert np.min(image.ivar ** -1) >= image.get_header_val('read_noise') ** (2)




import numpy as np
import copy
from astropy.table import Table

from xwavecal.utils import extract_utils
from xwavecal.stages import Stage


class BoxExtract(Stage):
    def __init__(self, runtime_context=None):
        super(BoxExtract, self).__init__(runtime_context=runtime_context)
        self.extraction_half_window = runtime_context.box_extraction_half_window
        self.max_extraction_half_window = runtime_context.max_extraction_half_window
        self.table_name = runtime_context.box_spectrum_name

    @staticmethod
    def extract_order(twod_spectrum, weights=None):
        """
        :param twod_spectrum: 2d float array.
                              array of flux where the center of the order is the central row.
        :param weights: 2d array
                        weights, the same shape as twod_spectrum
        :return: 1d array.
                 weighted sum along the columns of twod_spectrum.
        """
        if weights is None:
            return np.sum(twod_spectrum, axis=0)
        else:
            return np.sum(twod_spectrum * weights, axis=0)

    def extract(self, rectified_2d_spectrum, rectified_ivar):
        """
         :param rectified_2d_spectrum: Dictionary
                Dictionary where the keys are the trace id's from trace.get_id(),
                where rectified_2d_spectrum[trace_id]['val'] is a 2d float array (flux for the trace_id order).
                If half extraction window was 10, then rectified_2d_spectrum[trace_id]['flux']
                is 21 rows by 4096 columns (for a 4096 pixel wide image). One would sum this 2d
                spectrum along-columns to get a box extracted spectrum.
         :param rectified_ivar: Dictionary
                Same as rectified_2d_spectrum but ['val'] is the inverse variance.
        :return:
        """
        extracted_spectrum_per_order = {'id': [], 'flux': [], 'stderr': [], 'pixel': [],
                                        'fiber': [], 'ref_id': [], 'wavelength': []}
        for order_id in list(rectified_2d_spectrum.keys()):
            weights = self._weights(rectified_2d_spectrum[order_id]['val'], rectified_ivar[order_id]['val'])
            flux = self.extract_order(rectified_2d_spectrum[order_id]['val'], weights)
            stdvar = self.extract_order(np.power(rectified_ivar[order_id]['val'], -1,
                                                 where=~np.isclose(rectified_ivar[order_id]['val'], 0),
                                                 out=np.zeros_like(rectified_ivar[order_id]['val'])),
                                        safe_pow(weights, 2))
            extracted_spectrum_per_order['flux'].append(flux)
            extracted_spectrum_per_order['stderr'].append(np.sqrt(stdvar))
            extracted_spectrum_per_order['pixel'].append(np.arange(len(flux)))
            extracted_spectrum_per_order['id'].append(order_id)

            extracted_spectrum_per_order['wavelength'].append(np.ones_like(flux, dtype=float) * np.nan)
            extracted_spectrum_per_order['ref_id'].append(np.nan)
            extracted_spectrum_per_order['fiber'].append(np.nan)
        return Table(extracted_spectrum_per_order)

    def _weights(self, order_rect_spectrum, order_rect_ivar):
        return None

    def do_stage(self, image):
        self.logger.info('Extracting spectrum')
        rectified_2d_spectrum = self._trim_rectified_2d_spectrum(image.rectified_2d_spectrum)
        rectified_ivar = self._trim_rectified_2d_spectrum(image.rectified_ivar)
        spectrum = self.extract(rectified_2d_spectrum, rectified_ivar)
        image.data_tables[self.table_name] = Table(spectrum)
        return image

    def _trim_rectified_2d_spectrum(self, rectified_2d_spectrum):
        """
        :param rectified_2d_spectrum: Dictionary
               Dictionary where the keys are the trace id's from trace.get_id(),
               where rectified_2d_spectrum[trace_id]['val'] is a 2d float array (flux for the trace_id order).
               If half extraction window was 10, then rectified_2d_spectrum[trace_id]['val']
               is 21 rows by 4096 columns (for a 4096 pixel wide image). One would sum this 2d
               spectrum along-columns to get a box extracted spectrum.
        :return rectified_2d_spectrum: Dictionary
                 Same as input but trimmed so that each order's 2d spectrum only has 2 * extraction_half_window + 1 rows.
        NOTE: The output spectra per order have the center of the trace at the center of the spectrum. E.g. if
        extraction_half_window is 10, then the 2d spectra have 21 rows and the trace center (peak flux) lies at
        index 10 (indexing from 0).
        """
        trimmed_rectified_spectrum = copy.deepcopy(rectified_2d_spectrum)
        if self.extraction_half_window >= self.max_extraction_half_window:
            # short circuit
            self.logger.warning('Extraction window was chosen to be >= the max extraction window defined in settings.py.'
                           ' Defaulting to the max extraction window.')
            return rectified_2d_spectrum
        trim = self.max_extraction_half_window - self.extraction_half_window
        for order_id in list(rectified_2d_spectrum.keys()):
            for data_type in list(rectified_2d_spectrum[order_id].keys()):
                trimmed_rectified_spectrum[order_id][data_type] = rectified_2d_spectrum[order_id][data_type][trim:-trim]
        return trimmed_rectified_spectrum


class IVarExtract(BoxExtract):
    """
    Extraction with each pixel weighted by its inverse variance.
    """
    def __init__(self, runtime_context=None):
        super(IVarExtract, self).__init__(runtime_context=runtime_context)
        self.extraction_half_window = runtime_context.sne_extraction_half_window
        self.max_extraction_half_window = runtime_context.max_extraction_half_window
        self.table_name = runtime_context.ivar_spectrum_name

    def _weights(self, order_rect_spectrum, order_rect_ivar):
        normalization = self.extract_order(order_rect_ivar)
        return np.divide(order_rect_ivar, normalization, out=np.zeros_like(order_rect_ivar),
                         where=~np.isclose(normalization, 0))


class BlazeCorrectedExtract(IVarExtract):
    """
    Same as IVarExtract, meant to run after dividing the 2d spectrum by the blaze AND re-rectifying the spectrum.
    """
    def __init__(self, runtime_context=None):
        super(BlazeCorrectedExtract, self).__init__(runtime_context=runtime_context)
        self.extraction_half_window = runtime_context.sne_extraction_half_window
        self.max_extraction_half_window = runtime_context.max_extraction_half_window
        self.table_name = runtime_context.blaze_corrected_spectrum_name


class RectifyTwodSpectrum(Stage):
    def __init__(self, runtime_context=None):
        super(RectifyTwodSpectrum, self).__init__(runtime_context=runtime_context)
        self.max_extraction_half_window = runtime_context.max_extraction_half_window

    def do_stage(self, image):
        self.logger.info('Rectifying the 2d spectrum')
        if image.trace is None:
            self.logger.error('Image has empty trace attribute. Aborting extraction.')
            image.is_bad = True
            image.rectified_2d_spectrum = {}
            image.rectified_ivar = {}
            return image
        rectified_2d_spectrum = extract_utils.rectify_orders(image.data, image.trace,
                                                             half_window=self.max_extraction_half_window)
        image.rectified_2d_spectrum = rectified_2d_spectrum
        image.ivar = np.full_like(image.data, np.nan, dtype=np.double) if image.ivar is None else image.ivar
        image.rectified_ivar = extract_utils.rectify_orders(image.ivar, image.trace,
                                                            half_window=self.max_extraction_half_window,
                                                            nullify_mapped_values=False)
        return image


def safe_pow(arr, power):
    if arr is None:
        return arr
    return np.power(arr, power)



"""
Module for any functions which are instrument specific and specified in the config files.
E.g. parse_region_keyword will parse the overscan and data region values in the .fits files
and return a tuple of splice objects specifying either the overscan or data region respectively.
"""

import numpy as np


def parse_nres_region_keyword(key, index_from_one=True):
    """
    :param key: string of the form '[y1:y2],[x1:x2]'
    :return: tuple
             (slice(y1,y2,None), slice(x1,x2,None))
             or if index_from_one:
             (slice(y1-1,y2,None), slice(x1-1,x2,None))
    """
    boundaries = [np.array(i.split(':')).astype(int) for i in key.replace('[', '').replace(']', '').split(',')]
    if index_from_one:
        boundaries[0][0], boundaries[1][0] = boundaries[0][0] - 1, boundaries[1][0] - 1
    return tuple(slice(*boundary) for boundary in boundaries)


def parse_harps_region_keyword(key):
    # WARN this function assumes all HARPS frames have overscans in x only (and zero overscan in y).
    """
    :param key: tuple of the form (rx, ry, osx, osy, psx, psy, Nx, Ny)
    where each entry is an integer.
    osx and osy are the number of overscan pixels in x and y (not the extent, the number of pixels)
    osx and osy are the number of prescan pixels in x and y
    Nx, Ny are the full size of the image (i.e. NAXIS1 and NAXIS2)
    rx and ry are the number of region pixels in x and y. E.g. for the overscan region, rx, ry will equal osx, osy.
    :return: tuple
             (slice(y1,y2,None), slice(x1,x2,None))

             the rows (y1:y2) and columns (x1:x2) corresponding to the input region Nx, Ny.
             Note x2!= Nx because of how HARPS frames store the overscan information.
    """
    nx, ny, osx, osy, psx, psy, Nx, Ny = key
    if np.isclose(nx, Nx, atol=300):
        # we are processing the data region
        return slice(psy, psy + ny), slice(psx, psx + nx)
    else:
        # we are processing the overscan region (because we do not subtract by the prescan anywhere in this pipeline.
        return slice(0, Ny), slice(Nx - osx, Nx)



import mock
import numpy as np
import pytest
from astropy.table import Table

from xwavecal.utils.fiber_utils import fiber_states_from_header, fibers_state_to_filename, \
                                          lit_fibers, lit_wavecal_fibers, wavecal_fibers_from_header
from xwavecal.fibers import IdentifyFibers, MakeFiberTemplate
from xwavecal.images import Image, DataProduct

from xwavecal.tests.utils import FakeImage, FakeContext


def test_creation_from_header():
    assert (1, 1, 0) == fiber_states_from_header('targ&ThAr&none')
    assert (0, 1, 1) == fiber_states_from_header('none&ThAr&targ')
    assert (0, 1, 0) == fiber_states_from_header('none&ThAr&none')


def test_wavecal_fibers_from_header():
    assert (0, 1, 0) == wavecal_fibers_from_header('targ&ThAr&none', 'thar')
    assert (0, 1, 1) == wavecal_fibers_from_header('none&ThAr&ThAr', 'thar')


def test_fiber_state_to_filename():
    image = Image(header={'fiber_state': 'tung&tung&none'})
    assert fibers_state_to_filename(image) == '110'


class TestIdentifyFibers:
    CONTEXT = FakeContext()

    def test_build_fiber_column(self):
        wavecal_fibers = np.array([1, 2])
        fibers = np.array([0, 1, 2])
        num_traces = 7
        for matched_ids in [[2], [3], [4]]:
            assert IdentifyFibers.build_fiber_column(matched_ids, fibers, wavecal_fibers, num_traces)[matched_ids[0]] == 1
            assert IdentifyFibers.build_fiber_column(matched_ids, fibers, wavecal_fibers, num_traces,
                                                     low_fiber_first=False)[matched_ids[0]] == 2

        assert np.allclose(IdentifyFibers.build_fiber_column([3], fibers, wavecal_fibers, num_traces),
                           [1, 2, 0, 1, 2, 0, 1])
        assert np.allclose(IdentifyFibers.build_fiber_column([3], fibers, wavecal_fibers, num_traces, low_fiber_first=False),
                           [2, 1, 0, 2, 1, 0, 2])

    def test_build_fiber_column_multi_match(self):
        wavecal_fibers = np.array([1, 2])
        fibers = np.array([0, 1, 2])
        num_traces = 7
        expected_fiber_designations = [1, 2, 0, 1, 2, 0, 1]

        assert np.allclose(IdentifyFibers.build_fiber_column([3, 4], fibers, wavecal_fibers, num_traces),
                           expected_fiber_designations)
        assert np.allclose(IdentifyFibers.build_fiber_column([4, 3], fibers, wavecal_fibers, num_traces),
                           expected_fiber_designations)

    def test_build_fiber_column_single_fiber(self):
        wavecal_fibers = np.array([1])
        fibers = np.array([1])
        num_traces = 7
        for matched_ids in [[2], [3], [4]]:
            assert IdentifyFibers.build_fiber_column(matched_ids, fibers, wavecal_fibers, num_traces)[matched_ids[0]] == 1
            assert IdentifyFibers.build_fiber_column(matched_ids, fibers, wavecal_fibers, num_traces,
                                                     low_fiber_first=False)[matched_ids[0]] == 1

        assert np.allclose(IdentifyFibers.build_fiber_column([3], fibers, wavecal_fibers, num_traces), np.ones(7))
        assert np.allclose(IdentifyFibers.build_fiber_column([3], fibers, wavecal_fibers, num_traces, low_fiber_first=False), np.ones(7))

    def test_build_ref_id_column(self):
        fiber_ids = np.array([1, 2, 1, 2, 1])
        ref_ids = IdentifyFibers.build_ref_id_column([2, 3], fiber_ids, ref_id=50)
        assert np.allclose(ref_ids, [49, 49, 50, 50, 51])
        fiber_ids = np.array([0, 1, 2, 0, 1])
        ref_ids = IdentifyFibers.build_ref_id_column([1, 2], fiber_ids, ref_id=50)
        assert np.allclose(ref_ids, [50, 50, 50, 51, 51])
        fiber_ids = np.array([0, 2, 1, 0, 2, 1])
        ref_ids = IdentifyFibers.build_ref_id_column([1, 2], fiber_ids, ref_id=50, low_fiber_first=False)
        assert np.allclose(ref_ids, [49, 50, 50, 50, 51, 51])
        fiber_ids = np.array([0, 0, 0, 0, 0, 0])
        ref_ids = IdentifyFibers.build_ref_id_column([1], fiber_ids, ref_id=50, low_fiber_first=False)
        assert np.allclose(ref_ids, [49, 50, 51, 52, 53, 54])

    def test_build_ref_id_column_reversed_match_order(self):
        fiber_ids = np.array([1, 2, 1, 2, 1])
        ref_ids = IdentifyFibers.build_ref_id_column([3, 2], fiber_ids, ref_id=50)
        assert np.allclose(ref_ids, [49, 49, 50, 50, 51])

    def test_calibration_type(self):
        assert IdentifyFibers(self.CONTEXT).calibration_type == 'FIBERS'

    @mock.patch('xwavecal.fibers.IdentifyFibers.get_calibration_filename', return_value=None)
    @mock.patch('xwavecal.fibers.IdentifyFibers.on_missing_master_calibration')
    def test_do_stage_aborts_on_missing_cal(self, mock_warn, mock_cal):
        assert 'image' == IdentifyFibers(self.CONTEXT).do_stage(image='image')

    @mock.patch('xwavecal.fibers.IdentifyFibers.get_calibration_filename', return_value='')
    @mock.patch('os.path.exists', return_value=True)
    def test_do_stage_aborts_on_no_wavecal_fibers(self, mock_warn, mock_exists):
        image = FakeImage()
        image.data_tables[self.CONTEXT.main_spectrum_name] = {'flux': [1, 2]}
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 0, 0
        assert image == IdentifyFibers(self.CONTEXT).do_stage(image=image)

    @mock.patch('xwavecal.fibers.IdentifyFibers.get_calibration_filename', return_value='')
    @mock.patch('os.path.exists', return_value=True)
    def test_do_stage_aborts_on_length_zero_spectrum(self, mock_warn, mock_exists):
        image = FakeImage()
        image.data_tables[self.CONTEXT.main_spectrum_name] = {'flux': []}
        assert image == IdentifyFibers(FakeContext()).do_stage(image=image)

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('xwavecal.fibers.IdentifyFibers.apply_master_calibration')
    @mock.patch('xwavecal.fibers.IdentifyFibers.get_calibration_filename', return_value='/path/')
    def test_do_stage(self, fake_cal, mock_apply_cal, mock_os):
        IdentifyFibers().do_stage(image='image')
        mock_apply_cal.assert_called_with('image', '/path/')

    @pytest.mark.integration
    @mock.patch('xwavecal.images.fits.open')
    def test_apply_master_calibration(self, mock_load):
        image = FakeImage()
        context = FakeContext()
        context.ref_id = 1
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 1, 0
        image.fiber0_lit, image.fiber1_lit, image.fiber2_lit = 0, 1, 1
        image.set_header_val('read_noise', 0)
        spec = Table({'id': [0, 1, 2], 'flux': 100 * np.random.random((3, 30)),
                      'fiber': [1, 1, 1], 'ref_id': [0, 0, 0]})
        mock_load.return_value = {'fibers': DataProduct(data=spec[1])}  # fake fiber template
        image.data_tables = {context.main_spectrum_name: spec}
        image = IdentifyFibers(context).apply_master_calibration(image, '')
        spec = image.data_tables[context.main_spectrum_name]
        assert np.allclose(spec['ref_id'], [0, 1, 1])
        assert np.allclose(spec['fiber'], [2, 1, 2])

    @pytest.mark.integration
    @mock.patch('xwavecal.images.fits.open')
    def test_apply_master_calibration_single_fiber(self, mock_load):
        image = FakeImage()
        context = FakeContext()
        context.ref_id = 1
        image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 1, 0
        image.fiber0_lit, image.fiber1_lit, image.fiber2_lit = 0, 1, 0
        image.set_header_val('read_noise', 0)
        spec = Table({'id': [0, 1, 2], 'flux': 100 * np.random.random((3, 30)),
                      'fiber': [1, 1, 1], 'ref_id': [0, 0, 0]})
        mock_load.return_value = {'fibers': DataProduct(data=spec[1])}  # fake fiber template
        image.data_tables = {context.main_spectrum_name: spec}
        image = IdentifyFibers(context).apply_master_calibration(image, '')
        spec = image.data_tables[context.main_spectrum_name]
        assert np.allclose(spec['ref_id'], [0, 1, 2])
        assert np.allclose(spec['fiber'], [1, 1, 1])


class TestMakeFiberTemplate:
    def test_do_stage(self):
        context = FakeContext()
        context.template_trace_id = 2
        image = FakeImage()
        spec = Table({'id': [0, 1, 2, 3, 4], 'flux': 100 * np.random.random((5, 30))})
        image.data_tables = {context.main_spectrum_name: spec}
        image, template = MakeFiberTemplate(context).do_stage(image)
        assert template.get_header_val('type') == MakeFiberTemplate(context).calibration_type.lower()
        assert image.get_header_val('type') != MakeFiberTemplate(context).calibration_type.lower()
        assert np.allclose(spec[[0, 2, 4]]['flux'], template.data['flux'])


@mock.patch('xwavecal.images.fits.open')
def test_construct_single_fiber_template(mock_load):
    mock_load.return_value = {'fibers': DataProduct(data=Table({'flux': np.array([np.ones(5), np.ones(5)])}))}
    template = IdentifyFibers.construct_single_fiber_template(None, 'fibers', 2)
    assert np.allclose(template, np.array([np.ones(5), np.zeros(5), np.ones(5)]))
    mock_load.return_value = {'fibers': DataProduct(data=Table({'flux': np.zeros((3, 10))}))}
    template = IdentifyFibers.construct_single_fiber_template(None, 'fibers', 3)
    assert np.allclose(template.shape, (7, 10))
    template = IdentifyFibers.construct_single_fiber_template(None, 'fibers', 1)
    assert np.allclose(template.shape, (3, 10))


@mock.patch('xwavecal.fibers.correlate2d')
def test_identify_matching_orders(fake_correlate):
    fake_correlate.return_value = np.array([[0, 1, 0],
                                            [0, 1, 2],
                                            [0, 0, 0]])
    assert np.allclose(IdentifyFibers.identify_matching_orders(None, None, 2), [1, 0])


def test_get_lit_fibers():
    image = FakeImage()
    image.fiber0_lit, image.fiber1_lit, image.fiber2_lit = 1, 1, 0
    image.fiber0_wavecal, image.fiber1_wavecal, image.fiber2_wavecal = 0, 1, 1
    assert np.allclose(lit_fibers(image), [0, 1])
    assert np.allclose(lit_wavecal_fibers(image), [1, 2])




import numpy as np


def fiber_states_from_header(fiber_string):
    parsed_objects = fiber_string.split('&')
    fiber0_lit = parsed_objects[0].lower() != 'none'
    fiber1_lit = parsed_objects[1].lower() != 'none'
    fiber2_lit = parsed_objects[2].lower() != 'none'
    return int(fiber0_lit), int(fiber1_lit), int(fiber2_lit)


def wavecal_fibers_from_header(fiber_string, wavecal_key='thar'):
    parsed_objects = fiber_string.split('&')
    fiber0_wavecal = parsed_objects[0].lower() == wavecal_key
    fiber1_wavecal = parsed_objects[1].lower() == wavecal_key
    fiber2_wavecal = parsed_objects[2].lower() == wavecal_key
    return int(fiber0_wavecal), int(fiber1_wavecal), int(fiber2_wavecal)


def fibers_state_to_filename(image):
    return str(int(image.fiber0_lit)) + str(int(image.fiber1_lit)) + str(int(image.fiber2_lit))


def lit_wavecal_fibers(image):
    return np.array([0] * image.fiber0_wavecal + [1] * image.fiber1_wavecal + [2] * image.fiber2_wavecal)


def lit_fibers(image):
    return np.array([0] * image.fiber0_lit + [1] * image.fiber1_lit + [2] * image.fiber2_lit)



import numpy as np
from scipy.ndimage import map_coordinates


def rectify_orders(image_data, trace, half_window=10, debug=False, nullify_mapped_values=True):
    """
    :param image_data: xwavecal.images.Image.data
    :param trace: Trace object
    :param half_window: half of the size of the extraction window about the center of the traces. E.g. 10 means 21 pixel
                        window will be extracted.
    :param debug: boolean for returning an internal copy of the image which is used in extract.
    :param nullify_mapped_values: bool.
                                  True if the flux from a pixel can only be used in one order's spectrum.
                                  False if a pixel's flux can be double (or triple etc..) counted.
    :return rectified_orders: Dictionary
    Dictionary where the keys are the trace id's from trace.get_id(),
    where rectified_2d_spectrum[trace_id]['val'] is a 2d float array (flux for the trace_id order).
    If half extraction window was 10, then rectified_2d_spectrum[trace_id]['val']
    is 21 rows by 4096 columns (for a 4096 pixel wide image). One would sum this 2d
    spectrum along-columns to get a box extracted spectrum.
    """
    rectified_orders = {}
    num_orders = trace.num_traces_found()
    image_copy = np.copy(image_data)
    x_coordinates, y_coordinates = np.meshgrid(np.arange(image_data.shape[1]), np.arange(image_data.shape[0]))
    image_coordinates = {'x': x_coordinates, 'y': y_coordinates}
    for index in np.arange(num_orders):
        rectified_orders[trace.get_id(index)], image_copy = rectify_order(image_copy, image_coordinates,
                                                                          trace.get_centers(index),
                                                                          half_window=half_window,
                                                                          nullify_mapped_values=nullify_mapped_values)
    if debug:
        return rectified_orders, image_copy
    else:
        return rectified_orders


def rectify_order(image_data, image_coordinates, single_order_centers, half_window=10, nullify_mapped_values=True):
    """
    :param image_data: xwavecal.images.Image.data
    :param image_coordinates: dictionary with x and y coordinates for each pixel.
    :param single_order_centers: the y centers of the center of the rectification window
    :param half_window: the half width of the window about which one is rectifying the trace/order
    :param nullify_mapped_values: True if the flux from a pixel can only be used in one order's spectrum.
                                  False if a pixel's flux can be double (or triple etc..) counted.
    :return: An array of 2*half_window + 1 rows by width(image_data) about the order. The center of the order
             is right at the index of half_window, e.g. rectified_order[half_window] gives the flux down the
             center of the order.
    """
    spectrum_shape = (2*half_window + 1, image_data.shape[1])
    rectified_order = {'val': np.zeros(spectrum_shape),
                       'y': np.zeros(spectrum_shape),
                       'x': np.zeros(spectrum_shape)}
    x_coords = np.arange(image_data.shape[1])
    for offset, row in zip(np.arange(-half_window, half_window + 1), np.arange(spectrum_shape[0])):
        mapped_y_values = map_coordinates(image_coordinates['y'], [single_order_centers + offset, x_coords],
                                          order=0, mode='constant', cval=0, prefilter=False)
        rectified_order['val'][row] = image_data[(mapped_y_values, x_coords)]
        rectified_order['y'][row] = mapped_y_values
        rectified_order['x'][row] = x_coords
        if nullify_mapped_values:
            image_data[(mapped_y_values, x_coords)] = 0
    return rectified_order, image_data



import numpy as np

from xwavecal.utils.correlate import correlate2d


def test_correlate2d():
    arr = np.ones((5, 5))
    arr[1:4, 1:4] = 2
    sig = correlate2d(arr, 2 * np.ones((3, 3)), max_lag=1)
    assert sig[2, 2] == np.max(sig)


import numpy as np
import pytest

from xwavecal.tests.test_traces import FakeTraceImage
from xwavecal.tests.utils import fill_image_with_traces, FakeContext
from xwavecal.utils.trace_utils import Trace
from xwavecal.utils import extract_utils
from xwavecal.extract import BoxExtract, RectifyTwodSpectrum, IVarExtract, safe_pow


class TestRectify:
    CONTEXT = FakeContext()
    def test_rectify_orders(self):
        image = FakeTraceImage()
        hw = 10
        peak_intensity = 1E4
        image, trace_centers, second_order_coefficient_guess = fill_image_with_traces(image, poly_order_of_traces=4,
                                                                                      max_num_traces=3,
                                                                                      fiber_intensity=peak_intensity)
        trace = Trace(data={'id': np.arange(trace_centers.shape[0]), 'centers': trace_centers})
        rectified_orders, zeroed_image_data = extract_utils.rectify_orders(image.data, trace,
                                                                           half_window=hw,
                                                                           debug=True)
        assert np.allclose(zeroed_image_data, 0)
        assert not np.allclose(image.data, 0)
        for key, rect_spect in rectified_orders.items():
            assert np.isclose(np.median(rect_spect['val'][hw]), peak_intensity, rtol=0.02)

    def test_rectify_orders_assigns_coordinates_at_edge(self):
        trace = Trace(data={'id': [1], 'centers': [np.zeros(3)]})
        image_data = np.zeros((3, 3))
        hw = 1
        rectified_orders, zeroed_image_data = extract_utils.rectify_orders(image_data, trace,
                                                                           half_window=hw,
                                                                           debug=True)
        assert np.allclose(rectified_orders[1]['y'], np.array([[0, 0, 0],
                                                               [0, 0, 0],
                                                               [1, 1, 1]]))
        assert np.allclose(rectified_orders[1]['x'], np.array([[0, 1, 2],
                                                               [0, 1, 2],
                                                               [0, 1, 2]]))

    def test_rectify_orders_assigns_coordinates(self):
        trace = Trace(data={'id': [1], 'centers': [np.ones(3)]})
        image = FakeTraceImage()
        image.data = np.zeros((3, 3))
        hw = 1
        rectified_orders, zeroed_image_data = extract_utils.rectify_orders(image.data, trace,
                                                                           half_window=hw,
                                                                           debug=True)
        assert np.allclose(rectified_orders[1]['y'], np.array([[0, 0, 0],
                                                               [1, 1, 1],
                                                               [2, 2, 2]]))
        assert np.allclose(rectified_orders[1]['x'], np.array([[0, 1, 2],
                                                               [0, 1, 2],
                                                               [0, 1, 2]]))

    def test_rectify_curved_order_maps_all_values(self):
        image = FakeTraceImage()
        image, trace_centers, second_order_coefficient_guess = fill_image_with_traces(image, poly_order_of_traces=4,
                                                                                      max_num_traces=1)
        x_coordinates, y_coordinates = np.meshgrid(np.arange(image.data.shape[1]), np.arange(image.data.shape[0]))
        image_coordinates = {'x': x_coordinates, 'y': y_coordinates}
        single_order_centers = trace_centers[0]
        rectified_order, zeroed_image_data = extract_utils.rectify_order(image.data, image_coordinates,
                                                                         single_order_centers, half_window=10,
                                                                         nullify_mapped_values=True)
        assert np.allclose(zeroed_image_data, 0)

    def test_rectify_flat_order(self):
        hw = 10
        image = FakeTraceImage()
        image, trace_centers, second_order_coefficient_guess = fill_image_with_traces(image, poly_order_of_traces=1,
                                                                                      max_num_traces=1)
        x_coordinates, y_coordinates = np.meshgrid(np.arange(image.data.shape[1]), np.arange(image.data.shape[0]))
        image_coordinates = {'x': x_coordinates, 'y': y_coordinates}
        single_order_centers = trace_centers[0]
        rectified_order, image_data = extract_utils.rectify_order(image.data, image_coordinates,
                                                                  single_order_centers, half_window=hw,
                                                                  nullify_mapped_values=False)
        trace_y_value = int(trace_centers[0][0])
        assert np.allclose(rectified_order['val'], image_data[trace_y_value - hw: trace_y_value + hw + 1, :])

    def test_rectification_does_not_change_box_extract(self):
        image = FakeTraceImage()
        image, trace_centers, second_order_coefficient_guess = fill_image_with_traces(image, poly_order_of_traces=4,
                                                                                      max_num_traces=1)
        x_coordinates, y_coordinates = np.meshgrid(np.arange(image.data.shape[1]), np.arange(image.data.shape[0]))
        image_coordinates = {'x': x_coordinates, 'y': y_coordinates}
        single_order_centers = trace_centers[0]
        rectified_order, image_data = extract_utils.rectify_order(image.data, image_coordinates,
                                                                  single_order_centers, half_window=10,
                                                                  nullify_mapped_values=False)
        rectified_spectrum = BoxExtract(self.CONTEXT).extract_order(rectified_order['val'])
        spectrum = BoxExtract(self.CONTEXT).extract_order(image_data)
        assert np.allclose(spectrum / np.median(spectrum), 1)
        assert np.allclose(rectified_spectrum, spectrum)

    def test_empty_spectrum_on_missing_trace(self):
        image = FakeTraceImage()
        image.trace = None
        image = RectifyTwodSpectrum(self.CONTEXT).do_stage(image=image)
        assert image.rectified_2d_spectrum == {}


class TestBoxExtract:
    CONTEXT = FakeContext()
    def test_box_extract_accuracy(self):
        image = FakeTraceImage()
        image, trace_centers, second_order_coefficient_guess = fill_image_with_traces(image, poly_order_of_traces=4,
                                                                                      max_num_traces=1)
        spectrum = BoxExtract(self.CONTEXT).extract_order(image.data)
        assert np.allclose(spectrum / np.median(spectrum), 1)

    def test_box_extract_trims_rectified_2d_spectrum(self):
        max_extract_window = 10
        for half_window in [2, 6, 10, 15]:
            fake_spectrum = np.zeros((2 * max_extract_window + 1, 5))
            fake_spectrum[max_extract_window] = 1
            extractor = BoxExtract(self.CONTEXT)
            extractor.extraction_half_window = half_window
            extractor.max_extraction_half_window = max_extract_window
            trimmed_spectrum = extractor._trim_rectified_2d_spectrum(rectified_2d_spectrum={1: {'val': fake_spectrum}})
            hw = min(half_window, max_extract_window)
            assert np.isclose(trimmed_spectrum[1]['val'].shape[0], 2*hw + 1)
            assert np.allclose(trimmed_spectrum[1]['val'][hw], 1)

    @pytest.mark.integration
    def test_box_extract(self):
        image = FakeTraceImage()
        image, trace_centers, second_order_coefficient_guess = fill_image_with_traces(image, poly_order_of_traces=4,
                                                                                      max_num_traces=2,
                                                                                      fiber_intensity=1E4)
        image.trace = Trace(data={'id': np.arange(trace_centers.shape[0]), 'centers': trace_centers})
        image = RectifyTwodSpectrum(self.CONTEXT).do_stage(image)
        image = BoxExtract(self.CONTEXT).do_stage(image)
        for spectrum in image.data_tables[self.CONTEXT.box_spectrum_name]['flux']:
            assert np.median(spectrum) > 1E4


class TestIVarExtract:
    CONTEXT = FakeContext()

    def test_weights_are_normalized(self):
        w = IVarExtract(self.CONTEXT)._weights(None, np.random.random((10, 20)))
        assert np.allclose(w.shape, (10, 20))
        assert np.allclose(np.sum(w, axis=0), 1)

    def test_weights_handle_div_by_zero(self):
        w = IVarExtract(self.CONTEXT)._weights(None, np.zeros((10, 20)))
        assert np.allclose(w.shape, (10, 20))
        assert np.allclose(w, 0)


def test_safe_pow():
    power = np.random.randint(1, 4)
    assert safe_pow(None, power) is None
    assert np.isclose(safe_pow(5, power), np.power(5, power))



"""
traces.py: Module for finding xwavecal orders across a CCD.
"""
import os
from copy import deepcopy, copy

from xwavecal.utils.trace_utils import Trace, AllTraceFitter
from xwavecal.stages import Stage, ApplyCalibration


class TraceMaker(Stage):
    def __init__(self, runtime_context):
        super(TraceMaker, self).__init__(runtime_context)
        self.order_of_poly_fit = runtime_context.trace_fit_polynomial_order
        self.second_order_coefficient_guess = runtime_context.trace_fit_initial_degree_two_guess
        self.trace_table_name = runtime_context.trace_table_name
        self.xmin = runtime_context.window_for_trace_identification['min']
        self.xmax = runtime_context.window_for_trace_identification['max']
        self.min_peak_to_peak_spacing = runtime_context.min_fiber_to_fiber_spacing
        self.min_snr = runtime_context.min_snr_for_trace_identification

    @property
    def calibration_type(self):
        return 'TRACE'

    def do_stage(self, image):
        self.logger.info('fitting traces order by order', )
        fitter = AllTraceFitter(xmin=self.xmin, xmax=self.xmax,
                                min_peak_to_peak_spacing=self.min_peak_to_peak_spacing,
                                min_snr=self.min_snr)
        trace = Trace(data=None, filepath=copy(image.filepath), header=deepcopy(image.header), translator=image.translator,
                      num_centers_per_trace=image.data.shape[1], table_name=self.trace_table_name)
        trace = fitter.fit_traces(trace=trace, image_data=image.data,
                                  poly_fit_order=self.order_of_poly_fit,
                                  second_order_coefficient_guess=self.second_order_coefficient_guess,
                                  image_noise_estimate=image.get_header_val('read_noise'))
        trace.set_header_val('type', self.calibration_type.lower())
        trace.fiber0_lit, trace.fiber1_lit, trace.fiber2_lit = image.fiber0_lit, image.fiber1_lit, image.fiber2_lit
        self.logger.info('Created master trace')
        image.trace = trace
        return image, trace


class LoadTrace(ApplyCalibration):
    """
    Loads trace coefficients from file and appends them onto the image object.
    """
    def __init__(self, runtime_context):
        super(LoadTrace, self).__init__(runtime_context)

    @property
    def calibration_type(self):
        return 'TRACE'

    def apply_master_calibration(self, image, master_calibration_path):
        image.trace = Trace.load(master_calibration_path, extension_name=self.runtime_context.trace_table_name)
        master_trace_filename = os.path.basename(master_calibration_path)
        image.set_header_val('IDTRACE', (master_trace_filename, 'ID of trace centers file'))
        self.logger.info('Applied loaded trace centers to image')
        return image



import abc
import os

from xwavecal.database import query_db_for_nearest

import logging


class Stage(object):
    def __init__(self, runtime_context):
        self.runtime_context = runtime_context
        self.logger = logging.getLogger(self.__class__.__name__)

    @abc.abstractmethod
    def do_stage(self, image):
        return image  # pragma: no cover


class ApplyCalibration(Stage):
    def __init__(self, runtime_context):
        super(ApplyCalibration, self).__init__(runtime_context)

    @property
    def calibration_type(self):
        return 'None'  # pragma: no cover

    def do_stage(self, image):
        master_calibration_path = self.get_calibration_filename(image)
        if master_calibration_path is None:
            self.on_missing_master_calibration(image)
            return image
        elif not os.path.exists(master_calibration_path):
            self.on_missing_master_calibration(image)
            return image
        self.logger.info('Template/calibration loaded: {0}'.format(master_calibration_path))
        return self.apply_master_calibration(image, master_calibration_path)

    def on_missing_master_calibration(self, image):
        self.logger.error('Master calibration file not found.')

    def get_calibration_filename(self, image):
        return query_db_for_nearest(self.runtime_context.database_path,
                                    image, self.calibration_type.lower(),
                                    time_format=self.runtime_context.time_format)

    @abc.abstractmethod
    def apply_master_calibration(self, image, path):
        return image  # pragma: no cover



import numpy as np
from scipy import interpolate
from astropy.table import Table, Column
from xwavecal.utils.misc_utils import find_nearest
import numpy.polynomial.polynomial as poly
import itertools


class OverlapFitter:
    def fit(self, b_lines, r_lines, b_lines_flux, r_lines_flux, linear_scale_range, flux_tol=0.2, deg=2):
        """
        :param deg. int
               deg >= 2.
        :return: (x, xp), a tuple of 1d ndarrays of equal length. For a given index i, x[i] maps to xp[i] in wavelength.
                 I.e. x and xp are the set of overlapping pixels in wavelength space for this pair of diffraction orders.
        """
        coeffs = self._fit_overlap(b_lines, r_lines, b_lines_flux, r_lines_flux,
                                   linear_scale_range, pixel_tol=2, flux_tol=flux_tol, deg=deg)
        coeffs = self._refine_fit(coeffs, b_lines, r_lines, pixel_tol=10, deg=deg)
        red_peaks = match_peaks(b_lines, r_lines, coeffs, pixel_tol=1)
        return {'pixel': red_peaks, 'matched_pixel': coordinate_transform(red_peaks, coeffs), 'peaks': len(red_peaks)}

    def _fit_overlap(self, b_lines, r_lines_f, b_lines_flux, r_lines_flux_f, linear_scale_range, pixel_tol=2,
                     flux_tol=0.2, deg=2):
        best_coeffs = np.array([0, 1, 0])

        r_lines, r_lines_flux = np.copy(r_lines_f), np.copy(r_lines_flux_f)

        #r_lines = r_lines[np.argsort(r_lines_flux)[::-1]][:15]
        #r_lines_flux = np.sort(r_lines_flux)[::-1][:15]
        r_lines_flux = r_lines_flux[np.argsort(r_lines)][:15]
        r_lines = np.sort(r_lines)[:15]
        b_lines_flux = b_lines_flux[np.argsort(b_lines)][-20:]
        b_lines = np.sort(b_lines)[-20:]
        # build nearest function for finding matching blue peaks.
        nearest_blue = interpolate.interp1d(b_lines, b_lines, kind='nearest', bounds_error=False,
                                            fill_value=(np.min(b_lines), np.max(b_lines)))
        matches, clipped_arr1 = self.match_matrix(r_lines, b_lines, r_lines_flux, b_lines_flux,
                                                  tol=flux_tol)
        vander = poly.polyvander(clipped_arr1, deg=deg)
        peak_combinations = itertools.combinations(np.arange(len(clipped_arr1)), deg+2)

        match_count = []
        all_coeffs = []
        for indices in peak_combinations:
            vander_inv = np.linalg.pinv(vander[list(indices)])
            lines2 = self._blue_matches(b_lines, matches, indices)
            potential_matched_blue_lines = self.line_combinations(lines2)
            if len(potential_matched_blue_lines) > 0:
                coeffs = np.matmul(vander_inv, potential_matched_blue_lines)
                coeffs[:, 2:, 0] = np.abs(coeffs[:, 2:, 0])  # ensures higher order coeffs are positive and thus
                # that the mappings are one-to-one.
                mapped_lines = np.matmul(vander, coeffs)
                num_matched_peaks = np.count_nonzero(np.isclose(mapped_lines, nearest_blue(mapped_lines),
                                                                atol=pixel_tol), axis=1).flatten()
                in_range = np.logical_and(coeffs[:, 1, 0] > min(linear_scale_range),
                                          coeffs[:, 1, 0] < max(linear_scale_range))
                coeffs, num_matched_peaks = coeffs[in_range], num_matched_peaks[in_range]
                if len(num_matched_peaks) > 0:
                    match_count.append(np.max(num_matched_peaks))
                    all_coeffs.append(coeffs[np.argmax(num_matched_peaks)])
        if len(match_count) > 0:
            best_coeffs = all_coeffs[np.argmax(match_count)]
        return np.array(best_coeffs).flatten()

    def _refine_fit(self, coeffs, b_lines, r_lines, pixel_tol=10, deg=2):
        """
        :param coeffs: array-like.
                       array of coefficients [a, b, c, ...] for g(x) = a + b*x + c*x^2 etc..
        :param b_lines: array-like
                        x-locations (pixel-locations) of emission lines from the blue order.
        :param r_lines: array-like
                        x-locations (pixel-locations) of emission lines from the red order.
        :param pixel_tol: tolerance for g(xr) - xb for the peaks at xr and xb to be considered
                          a matched-peak (i.e. duplicates of one-another). Where xr is the pixel
                          coordinate of the peak from the red order, and xb is the pixel coordinate
                          of the peak from the blue order.
        :param deg: polynomial degree of the mapping g(x)
        :return: ndarray
                 coefficients for g(x), [a, b, c,...] which minimize the chi^2 of (g(xr) - xb) for all
                 matched peaks (xr, xb) within the overlap.
        """
        peaks = poly.polyval(r_lines, coeffs)
        blue_peaks = interpolate.interp1d(b_lines, b_lines, kind='nearest', bounds_error=False,
                                          fill_value=(np.min(b_lines), np.max(b_lines)))(peaks)
        if np.count_nonzero(np.isclose(blue_peaks, peaks, atol=pixel_tol)) >= deg + 1:
            coeffs = poly.polyfit(r_lines[np.isclose(blue_peaks, peaks, atol=pixel_tol)],
                                  blue_peaks[np.isclose(blue_peaks, peaks, atol=pixel_tol)], deg)
        return coeffs

    @staticmethod
    def _blue_matches(blue_lines, match_matrix, red_index):
        return [blue_lines[match_matrix[list(red_index)][i]] for i in range(len(red_index))]

    @staticmethod
    def line_combinations(iterable):
        """
        :param iterable: list.
               list of lists/ndarrays of x positions. E.g. [np.array([...]), np.array([...])...]
        :return: ndarray.
        all possible N-length combinations of the N arrays inside iterable, picking one element
        from each, subject to the constraint that each N-length combination is strictly increasing.
        Each combination is returned as a row vector. E.g. if iterable = [[1,5], [2, 6]],
        the return would be:
        [np.array([[1], [2]]), np.array([[1], [6]]), np.array([[5], [6]])]
        """
        paths = np.array(list(itertools.product(*iterable)))
        paths = paths[np.all(paths[:, :-1] < paths[:, 1:], axis=1)]
        return paths.reshape(-1, len(iterable), 1)

    @staticmethod
    def match_matrix(lines1, lines2, lines1_flux, lines2_flux, tol=0.2):
        # make boolean matrix which indicates where peak fluxes are close.
        flux1_matrix = lines1_flux.reshape(-1, 1) * np.ones((len(lines1), len(lines2)))
        flux2_matrix = np.ones((len(lines1), len(lines2))) * lines2_flux
        # check that the difference of each peak is less than tol times their mean value (representative of the `true` peak value)
        matches = np.abs(flux2_matrix - flux1_matrix) < tol * np.mean([flux2_matrix, flux1_matrix], axis=0)
        # do a cut on arr1 where arr2 fluxes do not match close enough
        # consider cutting elements of arr2 as well which have no flux matches.
        valid_lines1 = lines1[np.sum(matches, axis=1) > 0]
        matches = matches[np.sum(matches, axis=1) > 0]
        return matches, valid_lines1


def blank_overlap_table(max_overlap_red):
    t = Table([Column(name='ref_id', dtype=int),
               Column(name='fiber', dtype=int),
               Column(name='matched_ref_id', dtype=int),
               Column(name='pixel', shape=(max_overlap_red,)),
               Column(name='matched_pixel', shape=(max_overlap_red,)),
               Column(name='peaks'),
               Column(name='good', dtype=bool)])
    return t


def _select_lines(lines, spec, ridx, bidx, max_overlap_red, max_overlap_blue):
    r_select = np.logical_and(lines['order'] == spec['ref_id'][ridx],
                              lines['pixel'] < np.max(spec['pixel'][ridx][:max_overlap_red]))
    b_select = np.logical_and(lines['order'] == spec['ref_id'][bidx],
                              lines['pixel'] > np.min(spec['pixel'][bidx][-max_overlap_blue:]))
    fluxkey = 'corrected_flux' if lines.get('corrected_flux', None) is not None else 'flux'
    return lines['pixel'][r_select], lines['pixel'][b_select], lines[fluxkey][r_select], lines[fluxkey][b_select]


def fit_overlaps(spec, lines, max_overlap_red=1000, max_overlap_blue=2000, linear_scale_range=(0.5, 2),
                 fiber=np.nan, deg=2, flux_tol=0.2):
    """
    :param spec: astropy.table.Table : sf_spec['flux'][i] should be a 1d ndarray with flux for the
                    diffraction order labelled sf_spec['ref_id'][i].
    :param lines: astropy.table.Table : lines['flux'][i], lines['order'][i], lines['pixel'][i]
                  must be the flux, order id, and pixel of an emission line.
    :param max_overlap_red: int. the pixel size of the blue edge of the redder order to consider for overlap
    :param max_overlap_blue: int. the pixel size of the red edge of the bluer order to consider for overlap
    :param linear_scale_range: ordered tuple. The range of linear coefficients to consider for the mapping between
    the red order and the blue order.
    :param flux_tol: float.
                     float between (0, 1]. For two lines to be considered a matched peak, their fluxes must
                     agree to within flux_tol.
    :return: A table of overlaps.

    WARNING: sf_spec[i] must be redder than sf_spec[i+1] if __ is True.
    WARNING: The spectrum of each order must be contiguous and evenly
    sampled over pixel space.
    WARNING: Each order must go from blue to red proceeding from left to right (pixel 0 to pixel 4096)
    """
    max_overlap_red = min(max_overlap_red, spec['pixel'].data.shape[1])
    max_overlap_blue = min(max_overlap_blue, spec['pixel'].data.shape[1])
    overlaps = blank_overlap_table(max_overlap_red)
    fitter = OverlapFitter()
    for row in range(len(spec['flux']) - 1):
        ridx, bidx = row, row+1
        r_lines, b_lines, r_line_flux, b_line_flux = _select_lines(lines, spec, ridx,
                                                                   bidx, max_overlap_red,
                                                                   max_overlap_blue)
        if len(r_lines) < deg + 1 or len(b_lines) < deg + 1:
            # skip to next if there are not enough lines to constrain the fit.
            continue
        overlap = fitter.fit(b_lines, r_lines, b_line_flux, r_line_flux, linear_scale_range,
                             deg=deg, flux_tol=flux_tol)
        pad = max_overlap_red - len(overlap['pixel'])
        overlaps.add_row({'ref_id': spec['ref_id'][ridx], 'matched_ref_id': spec['ref_id'][bidx],
                          'pixel': np.pad(overlap['pixel'].astype(float), ((0, pad)),
                                          mode='constant', constant_values=np.nan),
                          'matched_pixel': np.pad(overlap['matched_pixel'].astype(float), ((0, pad)),
                                                  mode='constant', constant_values=np.nan),
                          'peaks': overlap['peaks'],
                          'good': np.ones_like(overlap['peaks'], dtype=bool),
                          'fiber': fiber * np.ones_like(overlap['peaks'], dtype=int)})
    return overlaps


def match_peaks(b_lines, r_lines, coeffs, pixel_tol=1):
    mapped_red_lines = coordinate_transform(r_lines, coeffs)
    return r_lines[np.isclose(mapped_red_lines, find_nearest(mapped_red_lines, b_lines), atol=pixel_tol)]


def coordinate_transform(coordinates, coefficients):
    return poly.polyval(x=coordinates, c=coefficients)


def flag_bad_overlaps(overlaps, min_num_matches):
    """
    :param overlaps: Table of overlaps with ref_id, matched_ref_id, pixel and
    matched_pixel columns
    :param min_num_matches: int. Minimum number of peak matches for an overlap to count as well fit.
    :return overlaps: Same is input but each bad overlap has had its entry in the 'good' column
                      set to False.
    """
    bad_overlaps = [idx for idx in range(len(overlaps)) if overlaps[idx]['peaks'] < min_num_matches]
    overlaps['good'][bad_overlaps] = False
    return overlaps


def flag_outlier_overlaps(overlaps):
    med = np.median([overlap['ref_id'] for overlap in overlaps if overlap['good']])
    outlier_overlaps = [idx for idx in range(len(overlaps)) if
                        overlaps[idx]['good'] and not np.isclose(overlaps[idx]['ref_id'], med, atol=10)]
    overlaps['good'][outlier_overlaps] = False
    return overlaps




import numpy as np
import copy
from scipy import optimize
from astropy.stats import median_absolute_deviation
from astropy.table import Column, Table, vstack
from numpy.polynomial import legendre

from scipy import interpolate

from xwavecal.images import Image
from xwavecal.stages import Stage, ApplyCalibration
from xwavecal.utils.wavelength_utils import identify_lines, calc_residuals, restrict, Model, _sigma_clip
from xwavecal.utils.wavelength_utils import estimate_global_scale, normalize_coordinates, pixel_order_as_array
from xwavecal.utils.overlap_utils import flag_bad_overlaps, fit_overlaps, blank_overlap_table, flag_outlier_overlaps
from xwavecal.utils.misc_utils import brute_local_min, find_nearest, minmax
from xwavecal.utils.fiber_utils import lit_wavecal_fibers

import logging
logger = logging.getLogger(__name__)


class WavelengthSolution(object):
    """
    Class to evaluate, contain and update the wavelength solution.

    Notable attributes:
        WavelengthSolution.model gives the current model of the wavelength solution.
        WavelengthSolution.model_coefficients gives the coefficients for the current model, in the order
        of WavelengthSolution.model.
    """
    def __init__(self, model=None, model_coefficients=None, measured_lines: dict = None, reference_lines=None,
                 m0=30, max_pixel=100, max_order=2, min_order=0, min_pixel=0, overlap_range=(),
                 grating_eq=True):
        if model is None:
            model = {1: [0, 1],
                     2: [0, 1]}
        self.max_order = max_order
        self.max_pixel = max_pixel
        self.min_pixel = min_pixel
        self.min_order = min_order
        self.model = Model(model)
        self.m0 = m0
        self.overlap_range = overlap_range
        self.model_coefficients = model_coefficients
        self.reference_lines = reference_lines
        self.grating_eq = grating_eq  # True to use a 1/(m0+i) prefactor.
        self._measured_lines = None
        self.measured_lines = measured_lines

    @property
    def measured_lines(self):
        return self._measured_lines

    @measured_lines.setter
    def measured_lines(self, lines):
        if lines is not None and lines.get('order', None) is not None and lines.get('pixel', None) is not None:
            lines['normed_pixel'] = normalize_coordinates(lines['pixel'], max_value=self.max_pixel, min_value=self.min_pixel)
            lines['normed_order'] = normalize_coordinates(lines['order'], max_value=self.max_order, min_value=self.min_order)
        self._measured_lines = lines

    def __call__(self, pixel, order):
        """
        :param pixel: ndarray. array (any shape) of pixel coordinates of spectral features.
        :param order: ndarray. Same length as pixel. array of order indices of spectral features.
        :return: ndarray. array of wavelengths. Same shape as pixel and order.
        """
        normed_pixel = normalize_coordinates(pixel, max_value=self.max_pixel, min_value=self.min_pixel)
        normed_order = normalize_coordinates(order, max_value=self.max_order, min_value=self.min_order)
        return self.wavelength_normed_input(normed_pixel, normed_order, order)

    def wavelength_normed_input(self, normed_pixel, normed_order, order, **kwargs):
        A, c = self._construct_wavelength_map_matrices(normed_pixel.flatten(), normed_order.flatten(), order.flatten())
        wavelengths = self._map_to_wavelength(A, c, self.model_coefficients)
        return wavelengths.reshape(normed_pixel.shape)

    def update_model(self, new_model, grating_eq=True):
        pixel, order = np.meshgrid(np.linspace(self.min_pixel, self.max_pixel, 30),
                                   np.arange(self.min_order, self.max_order + 1))
        coords = {'pixel': pixel.flatten(), 'order': order.flatten()}
        coords['normed_pixel'] = normalize_coordinates(coords['pixel'], self.max_pixel, self.min_pixel)
        coords['normed_order'] = normalize_coordinates(coords['order'], self.max_order, self.min_order)
        old_wavelengths = self.wavelength_normed_input(**coords)
        self.model = Model(new_model)
        self.grating_eq = grating_eq
        self.model_coefficients = self.solve(coords, old_wavelengths)

    def solve(self, line_coordinates, wavelengths_to_fit, weights=None):
        """
        :param weights: array.
                        A 1d array of weights. E.g. the square root of the inverse variance.
        :return: array: best fit coefficients
        """
        weights = np.array([1]) if weights is None else weights
        A, c = self._construct_wavelength_map_matrices(**line_coordinates)
        c += np.array(wavelengths_to_fit).reshape(-1, 1)
        model_coefficients, residuals = np.linalg.lstsq(A * weights.reshape(-1, 1),
                                                        c * weights.reshape(-1, 1), rcond=None)[:2]
        return model_coefficients.flatten()

    def solve_from_overlaps(self, overlaps):
        if 0 in self.model:
            logger.error('Model contains parameters independent of pixel coordinate x. '
                         'Overlap fit will fail. Do not include 0: [...] terms in '
                         'the initial wavelength model. The model is {0}'.format(self.model))
            raise ValueError('0 in self.model.keys(). Not allowed for overlap fit.')
        coordinates = self._format_overlaps(overlaps, pixel_key='pixel', order_key='ref_id')
        matched_coordinates = self._format_overlaps(overlaps, pixel_key='matched_pixel', order_key='matched_ref_id')
        A1, c1 = self._construct_wavelength_map_matrices(**coordinates)
        A2, c2 = self._construct_wavelength_map_matrices(**matched_coordinates)
        A, c = A1 - A2, c1 - c2
        model_coefficients = np.linalg.lstsq(A, c, rcond=None)[0]
        self.model_coefficients = model_coefficients.flatten()

    def apply_scale(self, scale):
        self.model_coefficients *= scale

    def _format_overlaps(self, overlaps, pixel_key='pixel', order_key='ref_id'):
        """
        :param overlaps: An astropy table of overlaps with the 'pixel' and 'matched_pixel' columns
        giving 1d arrays of pixel points which must map to the same wavelength.
        :return: coordinates of overlaps.
        """
        orders = (overlaps[order_key].data.reshape(-1, 1) * np.ones_like(overlaps[pixel_key].data)).flatten()
        pixels = overlaps[pixel_key].data.flatten()
        pixels, orders = pixels[np.isfinite(pixels)], orders[np.isfinite(pixels)]
        coordinates = {'normed_pixel': normalize_coordinates(pixels, self.max_pixel, self.min_pixel), 'pixel': pixels,
                       'normed_order': normalize_coordinates(orders, self.max_order, self.min_order), 'order': orders}
        return coordinates

    def _construct_wavelength_map_matrices(self, normed_pixel, normed_order, order, **kwargs):
        if self.grating_eq:
            m_divisor = 1 / np.add(order, self.m0).astype(np.float64)
        else:
            m_divisor = np.ones_like(normed_pixel, dtype=np.float64)
        columns = ()
        for xpower in self.model:
            for orderpower in self.model[xpower]:
                coefficient_array = np.zeros((xpower+1, orderpower+1))
                coefficient_array[xpower, orderpower] = 1
                columns += (m_divisor * legendre.legval2d(normed_pixel, normed_order, coefficient_array),)
        A = np.column_stack(columns)
        if (0 not in self.model) or (0 not in self.model[0]):
            # we force 0,0 coefficient to 1 for the overlap fit.
            c = -1 * m_divisor.reshape(-1, 1)
        else:
            c = np.zeros_like(normed_pixel, dtype=np.float64).reshape(-1, 1)
        return A, c

    @staticmethod
    def _map_to_wavelength(A, c, coefficients):
        if coefficients is None:
            return np.ones_like(c.flatten(), dtype=float) * np.nan
        return np.dot(A, coefficients).flatten() + (-1) * c.flatten()


class WavelengthStage(Stage):
    """
    Base class for all wavelength stages
    """
    def __init__(self, runtime_context=None):
        super(WavelengthStage, self).__init__(runtime_context=runtime_context)

    def do_stage(self, image):
        valid_fibers = self._valid_fibers(image)
        for fiber in valid_fibers:
            image = self.do_stage_fiber(image, fiber)
        if len(valid_fibers) < 1:
            self.on_no_valid_fibers(image)
        return image

    def do_stage_fiber(self, image, fiber):
        return image

    def on_no_valid_fibers(self, image):
        self.logger.warning('No fibers found with non None WavelengthSolution objects. Aborting this '
                            'stage.')

    @staticmethod
    def _valid_fibers(image):
        return [fiber for fiber in lit_wavecal_fibers(image) if
                image.wavelength_solution[fiber] is not None]


class Initialize(WavelengthStage):
    """
    Stage 1/9 for the wavelength solution
    """
    def __init__(self, runtime_context=None):
        super(Initialize, self).__init__(runtime_context=runtime_context)

    def do_stage_fiber(self, image, fiber):
        self.logger.info('Appending blank WavelengthSolution object to image for this fiber. '
                         'fiber={0}'.format(str(fiber)))
        spectrum = image.data_tables[self.runtime_context.main_spectrum_name]
        single_fiber_spectrum = spectrum[spectrum['fiber'] == fiber]
        image.wavelength_solution[fiber] = WavelengthSolution(model=self.runtime_context.initial_wavelength_model,
                                                              m0=self.runtime_context.principle_order_number,
                                                              max_order=np.max(single_fiber_spectrum['ref_id']),
                                                              min_order=np.min(single_fiber_spectrum['ref_id']),
                                                              max_pixel=np.max(single_fiber_spectrum['pixel']),
                                                              min_pixel=np.min(single_fiber_spectrum['pixel']))
        return image

    def on_no_valid_fibers(self, image):
        for fiber in lit_wavecal_fibers(image):
            image.wavelength_solution[fiber] = None
        self.logger.error('Aborting wavelength calibration and setting wavelength solution to None.')

    def _valid_fibers(self, image):
        if image.data_tables.get(self.runtime_context.main_spectrum_name) is None:
            self.logger.error('No main spectrum on image.')
            return []

        colnames_present = all([key in image.data_tables[self.runtime_context.main_spectrum_name].colnames
                                for key in ['ref_id', 'fiber']])
        if not colnames_present:
            self.logger.error('Main spectrum on image is missing ref_id and fiber columns')
            return []
        return lit_wavecal_fibers(image)


class LoadReferenceLineList(ApplyCalibration):
    """
    Stage 2/8 for the wavelength solution
    """
    def __init__(self, runtime_context=None):
        super(LoadReferenceLineList, self).__init__(runtime_context=runtime_context)

    def apply_master_calibration(self, image, reference_list_path):
        line_list = np.sort(np.genfromtxt(reference_list_path, usecols=[1]).flatten())
        for fiber in [fiber for fiber in lit_wavecal_fibers(image) if
                      image.wavelength_solution[fiber] is not None]:
            image.wavelength_solution[fiber].reference_lines = line_list
        image.set_header_val('IDLIST', (reference_list_path, 'ID of the reference line list'))
        return image

    def get_calibration_filename(self, image):
        return self.runtime_context.line_list_path


class FitOverlaps(WavelengthStage):
    """
    Stage 3/8 for the wavelength solution
    This should run on a blaze corrected calibration spectrum.
    """
    def __init__(self, runtime_context=None):
        super(FitOverlaps, self).__init__(runtime_context=runtime_context)

    def do_stage(self, image):
        if image.data_tables.get(self.runtime_context.overlap_table_name, None) is None:
            image.data_tables[self.runtime_context.overlap_table_name] = blank_overlap_table(self.runtime_context.max_red_overlap)

        image = super(FitOverlaps, self).do_stage(image)

        name = self.runtime_context.overlap_table_name
        image.data_tables[name] = Table(image.data_tables[name])
        return image

    def do_stage_fiber(self, image, fiber):
        self.logger.info('Fitting overlaps. fiber={0}'.format(str(fiber)))
        spectrum = image.data_tables[self.runtime_context.main_spectrum_name]
        single_fiber_spectrum = spectrum[spectrum['fiber'] == fiber]
        overlaps = fit_overlaps(spec=single_fiber_spectrum,
                                lines=image.wavelength_solution[fiber].measured_lines,
                                max_overlap_red=self.runtime_context.max_red_overlap,
                                max_overlap_blue=self.runtime_context.max_blue_overlap,
                                linear_scale_range=self.runtime_context.overlap_linear_scale_range,
                                fiber=fiber,
                                flux_tol=getattr(self.runtime_context, 'flux_tol', 0.2))
        overlaps = flag_bad_overlaps(overlaps, getattr(self.runtime_context, 'min_num_matches', 6))
        self.logger.info('{0} overlaps verified. fiber={1}'.format(np.count_nonzero(overlaps['good']), str(fiber)))
        overlaps = flag_outlier_overlaps(overlaps)
        self.logger.info('{0} overlaps will be used. fiber={1}'.format(np.count_nonzero(overlaps['good']), str(fiber)))

        image.data_tables[self.runtime_context.overlap_table_name] = vstack([overlaps,
                                                             image.data_tables[self.runtime_context.overlap_table_name]])
        if np.count_nonzero(overlaps['good']) < self.runtime_context.min_num_overlaps:
            self.logger.error('Less than {0} overlaps verified as good,'
                         'setting wavelength solution to None.'
                         ' fiber={1}'.format(self.runtime_context.min_num_overlaps, str(fiber)))
            image.wavelength_solution[fiber] = None
        return image


class SolveFromOverlaps(WavelengthStage):
    """
    Stage 4/8. Solves for the coefficients of the wavelength solution from the overlaps.
    """
    def __init__(self, runtime_context=None):
        super(SolveFromOverlaps, self).__init__(runtime_context=runtime_context)

    def do_stage_fiber(self, image, fiber):
        image.wavelength_solution[fiber].model = Model(self.runtime_context.initial_wavelength_model)
        overlaps = image.data_tables.get(self.runtime_context.overlap_table_name, blank_overlap_table(1))
        overlaps = self._prune_overlaps(overlaps, fiber)
        self.logger.info('Initializing wavelength solution from overlaps. fiber={0}'.format(str(fiber)))
        image.wavelength_solution[fiber].overlap_range = minmax([overlaps['ref_id'], overlaps['matched_ref_id']])
        image.wavelength_solution[fiber].solve_from_overlaps(overlaps)
        return image

    @staticmethod
    def _prune_overlaps(overlaps, fiber):
        overlaps = overlaps[overlaps['good']]
        overlaps = overlaps[overlaps['fiber'] == fiber]
        return overlaps


class IdentifyArcEmissionLines(WavelengthStage):
    """
    Stage 5/8 for the wavelength solution
    """
    def __init__(self, runtime_context=None):
        super(IdentifyArcEmissionLines, self).__init__(runtime_context=runtime_context)
        self.min_peak_snr = self.runtime_context.min_peak_snr

    def do_stage_fiber(self, image, fiber):
        spectrum = image.data_tables[self.runtime_context.main_spectrum_name]
        single_fiber_spectrum = spectrum[spectrum['fiber'] == fiber]
        measured_lines = identify_lines(spectrum=single_fiber_spectrum,
                                        stderr=single_fiber_spectrum['stderr'],
                                        min_snr=self.min_peak_snr,
                                        order_key='ref_id')
        image.wavelength_solution[fiber].measured_lines = measured_lines
        self.logger.info('{0} emission lines identified from {1} unique '
                    'diffraction orders. fiber={2}'
                    ''.format(len(measured_lines['pixel']), len(set(measured_lines['order'])), str(fiber)))
        return image


class BlazeCorrectArcEmissionLines(WavelengthStage):
    def __init__(self, runtime_context=None):
        super(BlazeCorrectArcEmissionLines, self).__init__(runtime_context=runtime_context)

    def do_stage_fiber(self, image, fiber):
        spectrum = image.data_tables.get(self.runtime_context.blaze_corrected_spectrum_name)
        lines = image.wavelength_solution[fiber].measured_lines
        if spectrum is None:
            image.wavelength_solution[fiber].measured_lines['corrected_flux'] = lines['flux']
        else:
            spectrum = spectrum[spectrum['fiber'] == fiber]
            lines['corrected_flux'] = np.zeros_like(lines['flux'], dtype=float)
            for spec in spectrum:
                flux = interpolate.interp1d(spec['pixel'], spec['flux'], kind='nearest', bounds_error=False, fill_value=0)
                in_order = np.where(lines['order'] == spec['ref_id'])
                lines['corrected_flux'][in_order] = flux(lines['pixel'][in_order])
            image.wavelength_solution[fiber].measured_lines = lines
        return image


class IdentifyArcEmissionLinesLowSN(IdentifyArcEmissionLines):
    def __init__(self, runtime_context=None):
        super(IdentifyArcEmissionLinesLowSN, self).__init__(runtime_context=runtime_context)
        self.min_peak_snr = self.runtime_context.overlap_min_peak_snr


class FindGlobalScale(WavelengthStage):
    """
    Stage 6/8 for the wavelength solution
    """
    def __init__(self, runtime_context=None):
        super(FindGlobalScale, self).__init__(runtime_context=runtime_context)
        self.step = getattr(self.runtime_context, 'global_scale_spacing', 10)

    def do_stage_fiber(self, image, fiber):
        scale_guess = estimate_global_scale(detector_range=self.runtime_context.approx_detector_range_angstroms,
                                            n=self.runtime_context.approx_num_orders,
                                            m0=image.wavelength_solution[fiber].m0)
        scale = self._find_scale(image.wavelength_solution[fiber], scale_guess, self.runtime_context.global_scale_range,
                                 step=self.step)
        image.wavelength_solution[fiber].update_model(self.runtime_context.intermediate_wavelength_model)
        image.wavelength_solution[fiber].apply_scale(scale)
        self.logger.info('The scale guess was {0:.6e} and the search yielded {1:.6e}. fiber={2}'
                    ''.format(scale_guess, scale, str(fiber)))
        if not np.isclose(scale, scale_guess, rtol=2):
            self.logger.error('Global scale is more than a factor of two away from initial guess, '
                         'an error in the wavelength solution for this fiber is likely. fiber={0}'.format(str(fiber)))
        return image

    @staticmethod
    def _find_scale(wcs, scale_guess, rrange=(0.1, 5), step=10):
        unscaled_wavelengths = wcs.wavelength_normed_input(**restrict(wcs.measured_lines, *wcs.overlap_range))
        args = (unscaled_wavelengths, np.sort(wcs.reference_lines))
        scale = optimize.minimize(fun=FindGlobalScale._chi_squared_safe, x0=scale_guess, args=args,
                                  method=brute_local_min, options={'step': step, 'rrange': rrange, 'filtw': 501,
                                                                   'finish': 'Nelder-Mead'}).x[0]
        return scale

    @staticmethod
    def _chi_squared_safe(scales, unscaled_wavelengths, reference_wavelengths, mem_limit=1E8):
        chi2 = np.zeros_like(scales, dtype=np.float32)
        step = max(int(len(scales) * len(unscaled_wavelengths)) // int(mem_limit * 8 / 32), 1)  # bits / bit_limit
        for i in range(step):
            chi2[i::step] = FindGlobalScale._chi_squared(scales[i::step], unscaled_wavelengths, reference_wavelengths)
        return chi2

    @staticmethod
    def _chi_squared(scales, unscaled_wavelengths, reference_wavelengths):
        wavelengths = np.array(scales).reshape(-1, 1) * unscaled_wavelengths
        residuals = calc_residuals(wavelengths.flatten(), reference_wavelengths).reshape(wavelengths.shape)
        return np.sum(residuals ** 2, axis=1).flatten()
        #return np.sum(np.sort(residuals**2, axis=1)[:, :nmax], axis=1).flatten()


class SolutionRefineInitial(WavelengthStage):
    """
    Stage 7/8 for the wavelength solution
    """
    def __init__(self, runtime_context=None):
        super(SolutionRefineInitial, self).__init__(runtime_context=runtime_context)
        self.kappa = getattr(runtime_context, 'initial_mad_clip', 6)

    def do_stage_fiber(self, image, fiber):
        image.wavelength_solution[fiber].update_model(new_model=self.runtime_context.intermediate_wavelength_model)
        image.wavelength_solution[fiber], rsd = self.constrain_solution_over_detector(image.wavelength_solution[fiber],
                                                                                      self.kappa)

        mad, std = median_absolute_deviation(rsd), np.std(rsd)
        self.logger.info('median absolute deviation is {0} and the standard deviation is {1}.'
                    ' fiber={2}'.format(mad, std, str(fiber)))
        self.logger.info('{0} lines within 4.5 median absolute deviations and {1} lines within 4.5 standard deviations'
                    ''.format(np.count_nonzero(np.isclose(rsd, 0, atol=4.5*mad)),
                              np.count_nonzero(np.isclose(rsd, 0, atol=4.5*std))))
        return image

    @staticmethod
    def constrain_solution_over_detector(wcs, kappa=6):
        """
        :param wcs: WavelengthSolution
        :param kappa: outliers with residuals exceeding kappa m.a.d. from 0 will not be used to constrain the solution.
        :return:     (WavelengthSolution, array)
        wavelength solution, residuals. Each entry of residuals is the difference in
        wavelength between a measured line and its corresponding closest reference line.
        """
        reference_lines = np.sort(wcs.reference_lines)
        residuals = []
        for order_range in SolutionRefineInitial._ranges_to_evaluate(*wcs.overlap_range, wcs.min_order, wcs.max_order):
            wcs, residuals = refine_wcs(wcs, wcs.measured_lines, reference_lines,
                                        SolutionRefineInitial._converged_when_max_iter,
                                        SolutionRefineInitial._clip,
                                        kwargs={'sigma': kappa,
                                                'stdfunc': median_absolute_deviation,
                                                'order_range': order_range},
                                        max_iter=5)
        return wcs, residuals

    @staticmethod
    def _converged_when_max_iter(*args, **kwargs):
        return False

    @staticmethod
    def _clip(wcs, lines, reference_lines, sigma=3, stdfunc=np.std, order_range=(), **kwargs):
        lines = restrict(lines, *order_range)
        residuals = calc_residuals(wcs.wavelength_normed_input(**lines), reference_lines)
        lines = _sigma_clip(residuals, lines, sigma=sigma, stdfunc=stdfunc)
        return lines

    @staticmethod
    def _ranges_to_evaluate(low, high, min_order, max_order):
        return [(low - i, high + i) for i in range(max(low - min_order, max_order - high) + 1)]


def refine_wcs(wcs, measured_lines, reference_lines, converged, clip_fun, kwargs=None, max_iter=20):
    """
    :param wcs: WavelengthSolution
    :param reference_lines: array. Wavelengths of reference calibration lines.
    :param measured_lines: dict: {'normed_pixel': array, 'normed_order': array}. Coordinates of measured lines.
    :param kwargs: keyword arguments to feed to clip_fun and converged.
    :param max_iter: the max number of iterations allowed to reach convergence.
    :param converged: function which accepts wcs, residuals, last_residuals, measured_lines
                      and **kwargs and returns True or False. If True, the solve loop will stop.
    :param clip_fun: function which accepts residuals, measured_lines, reference_lines and **kwargs and returns
    a dict in the same format as measured_lines.
    :return:     (WavelengthSolution, array)
    wavelength solution, residuals. Each entry of residuals is the difference in
    wavelength between a measured line and its corresponding closest reference line.
    """
    if kwargs is None:
        kwargs = {}
    residuals, last_residuals = [], np.inf
    # note we could evaluate the find_nearest function out here instead of in the loop.
    for iteration in range(max_iter):
        clipped_lines = clip_fun(wcs, measured_lines, reference_lines, **kwargs)
        closest_ref_lines = find_nearest(wcs.wavelength_normed_input(**clipped_lines), array_b=reference_lines)
        wcs.model_coefficients = wcs.solve(clipped_lines, closest_ref_lines, clipped_lines.get('weight', None))
        residuals = calc_residuals(wcs.wavelength_normed_input(**measured_lines), reference_lines)
        if converged(residuals, last_residuals, **kwargs):
            break
        last_residuals = np.copy(residuals)
    return wcs, residuals


class SolutionRefineFinal(WavelengthStage):
    """
    Stage 8/8 for the wavelength solution
    """
    def __init__(self, runtime_context=None):
        super(SolutionRefineFinal, self).__init__(runtime_context=runtime_context)
        self.kappa = getattr(runtime_context, 'final_mad_clip', 4)

    def do_stage_fiber(self, image, fiber):
        image.wavelength_solution[fiber], rsd = self._refine(image.wavelength_solution[fiber],
                                                             self.runtime_context.final_wavelength_model, self.kappa)

        mad, std = median_absolute_deviation(rsd), np.std(rsd)
        self.logger.info('median absolute deviation is {0} and the standard deviation is {1}.'
                    ' fiber={2}'.format(mad, std, str(fiber)))
        self.logger.info('{0} lines within 4.5 median absolute deviations and {1} lines within 4.5 standard deviations'
                    ''.format(np.count_nonzero(np.isclose(rsd, 0, atol=4.5*mad)),
                              np.count_nonzero(np.isclose(rsd, 0, atol=4.5*std))))
        return image

    @staticmethod
    def _refine(wavelength_solution, final_model, kappa=4):
        """
        :param wavelength_solution: WavelengthSolution
        :param final_model: dict. The model which describes the form of the mapping from pixel and order
                            to wavelength. See __init__ of WavelengthSolution.
        :param kappa: outliers with residuals exceeding kappa m.a.d. from 0 will not be used to constrain the solution.
        :return:     (WavelengthSolution, array)
        wavelength solution, residuals. Each entry of residuals is the difference in
        wavelength between a measured line and its corresponding closest reference line.
        """
        reference_lines = np.sort(wavelength_solution.reference_lines)
        measured_lines = wavelength_solution.measured_lines
        residuals = []
        for x_degree in final_model:
            for i_degree in final_model[x_degree]:
                if wavelength_solution.model.is_missing_polynomial_term(x_degree, i_degree):
                    # This will only refine if the model is missing the term,
                    #  which means this stage does nothing if you want to rerun a model.
                    new_model = copy.deepcopy(wavelength_solution.model)
                    new_model.add_polynomial_term(x_degree, i_degree)
                    wavelength_solution.update_model(new_model=new_model)
                    wavelength_solution, residuals = refine_wcs(wavelength_solution,
                                                                measured_lines, reference_lines,
                                                                SolutionRefineFinal._converged,
                                                                SolutionRefineFinal._clip,
                                                                max_iter=20,
                                                                kwargs={'sigma': kappa,
                                                                        'stdfunc': median_absolute_deviation})
        return wavelength_solution, residuals

    @staticmethod
    def _converged(residuals, last_residuals, **kwargs):
        if np.allclose(residuals, last_residuals, atol=1E-3):
            return True
        return False

    @staticmethod
    def _clip(wcs, lines, reference_lines, sigma=3, stdfunc=np.std, **kwargs):
        residuals = calc_residuals(wcs.wavelength_normed_input(**lines), reference_lines)
        lines = _sigma_clip(residuals, lines, sigma=sigma, stdfunc=stdfunc)
        return lines


class SolutionRefineOnce(SolutionRefineFinal):
    """
    Single iteration of refining the wavelength solution. Useful if more lines have been added
    to the line list after final refine.
    """
    def __init__(self, runtime_context=None):
        super(SolutionRefineOnce, self).__init__(runtime_context=runtime_context)

    def do_stage_fiber(self, image, fiber):
        image.wavelength_solution[fiber], rsd = refine_wcs(image.wavelength_solution[fiber],
                                                           image.wavelength_solution[fiber].measured_lines,
                                                           image.wavelength_solution[fiber].reference_lines,
                                                           self._converged, self._clip, max_iter=20,
                                                           kwargs={'sigma': 4,
                                                                   'stdfunc': median_absolute_deviation})

        mad, std = median_absolute_deviation(rsd), np.std(rsd)
        self.logger.info('median absolute deviation is {0} and the standard deviation is {1}.'
                    ' fiber={2}'.format(mad, std, str(fiber)))
        self.logger.info('{0} lines within 4.5 median absolute deviations and {1} lines within 4.5 standard deviations'
                    ''.format(np.count_nonzero(np.isclose(rsd, 0, atol=4.5*mad)),
                              np.count_nonzero(np.isclose(rsd, 0, atol=4.5*std))))
        return image


class ApplyToSpectrum(WavelengthStage):
    """
    Stage 8/9 for the wavelength solution
    """
    def __init__(self, runtime_context=None):
        super(ApplyToSpectrum, self).__init__(runtime_context=runtime_context)

    def do_stage_fiber(self, image, fiber):
        spectrum = image.data_tables[self.runtime_context.main_spectrum_name]
        fiber_mask = np.where(spectrum['fiber'] == fiber)
        wcs = image.wavelength_solution[fiber]
        pixel_coordinates, order_coordinates = pixel_order_as_array(spectrum[fiber_mask])
        spectrum['wavelength'][fiber_mask] = wcs(pixel=pixel_coordinates, order=order_coordinates)
        spectrum.meta['header'] = {'MODEL': str(wcs.model), 'MCOEFFS': str(list(wcs.model_coefficients))}
        image.data_tables[self.runtime_context.main_spectrum_name] = spectrum
        return image


class TabulateArcEmissionLines(WavelengthStage):
    """
    Optional stage which will create a table of all identified emission lines and their matches
    in the reference list. This is saved when the image reduction is completed, under the extension LINES.
    """
    def __init__(self, runtime_context=None):
        super(TabulateArcEmissionLines, self).__init__(runtime_context=runtime_context)

    def do_stage(self, image):
        # TODO if wavelength solution fails, then the emission lines are not saved. Make it so that they are
        #  always saved
        valid_fibers = self._valid_fibers(image)
        if len(valid_fibers) > 0:
            lines = self._format_lines(image, valid_fibers)
            for fiber in valid_fibers:
                wcs = image.wavelength_solution[fiber]
                fib = np.where(lines['fiber'] == fiber)
                lines['wavelength'][fib] = wcs(lines['pixel'][fib].data, lines['order'][fib].data)
                lines['reference_wavelength'][fib] = find_nearest(lines['wavelength'][fib], wcs.reference_lines)

            image.data_tables[self.runtime_context.emission_lines_table_name] = Table(lines)
        return image

    @staticmethod
    def _format_lines(image, fibers):
        all_lines = []
        for fiber in fibers:
            lines = Table(image.wavelength_solution[fiber].measured_lines)
            lines.add_column(Column(fiber * np.ones_like(lines['pixel']), dtype=int), name='fiber')
            lines.add_column(Column(np.zeros_like(lines['pixel'], dtype=np.float64), unit='angstrom'), name='wavelength')
            lines.add_column(Column(np.zeros_like(lines['pixel'], dtype=np.float64), unit='angstrom'), name='reference_wavelength')
            lines['reference_wavelength'].description = 'wavelength of the nearest match in the reference line list'
            lines['wavelength'].description = 'wavelength of measured line from the calibration frame'
            all_lines.append(lines)
        return vstack(all_lines)


class IdentifyPrincipleOrderNumber(WavelengthStage):
    """
    Stage for identifying the principle order number m0. This should be run between IdentifyArcEmissionLines and
    SolveFromOverlaps. This stage is meant to be run once on a batch of stages to find the principle order
    number. The principle order number should then be fixed in the config.ini file for your instrument.
    """
    def __init__(self, runtime_context=None):
        super(IdentifyPrincipleOrderNumber, self).__init__(runtime_context=runtime_context)
        self.STAGES_TODO = [SolveFromOverlaps, FindGlobalScale, SolutionRefineInitial, SolutionRefineFinal]

    def do_stage_fiber(self, image, fiber):
        self.logger.info('Looking for the principle order number between {0} and {1}.'
                         ' fiber={2}'.format(*self.runtime_context.m0_range, str(fiber)))
        self.logger.disabled = True
        merits, m0_values = self.merit_per_m0(image, fiber, self.runtime_context.m0_range)
        self.logger.disabled = False
        best_m0, merit = m0_values[np.argmin(merits)], np.min(merits)

        if not merit < 1/10 * np.median(merits):
            self.logger.warning('A definitive principle order number was not found. Aborting wavelength solution.'
                           ' fiber={0}'.format(str(fiber)))
            image.wavelength_solution[fiber] = None
        else:
            image.wavelength_solution[fiber].m0 = best_m0
            self.logger.info('The best principle order number is {0}. fiber={1}'.format(best_m0, str(fiber)))
        return image

    def merit_per_m0(self, image, fiber, m0_range):
        m0_values = np.arange(*m0_range)
        merit = []
        for m0 in m0_values:
            image.wavelength_solution[fiber].m0 = m0
            for STAGE in self.STAGES_TODO:
                image = STAGE(self.runtime_context).do_stage_fiber(image, fiber)

            wcs = image.wavelength_solution[fiber]
            if wcs is not None:
                merit.append(median_absolute_deviation(calc_residuals(wcs.wavelength_normed_input(**wcs.measured_lines),
                                                             wcs.reference_lines)))
        return np.array(merit), m0_values


def find_feature_wavelengths(measured_lines, reference_lines, m0_range: tuple, max_pixel, min_pixel=0,
                             wavelength_models: dict = None, overlap_settings: dict = None, scale_settings: dict = None,
                             stages_todo: list = None):
    """
    A convenience function for any users who want to be able to wavelength calibrate from a list of
    spectral feature (measured_lines) pixel and order positions and a reference line list (reference_lines).
    This is not used anywhere in xwavecal, nor the included pipeline, except in
    xwavecal.tests.test_wavelength.

    The measured_lines should be for one fiber only. This function should be looped over fibers to calibrate
    many fibers.

    This function was designed to be used in Banzai-NRES. It returns the wavelengths of the spectral features
    input in measured_lines.

    :param measured_lines: dict.
           dictionary of ndarrays of pixel and order positions of measured spectral features (e.g. emission lines).
           Example:
               measured_lines = {'pixel': np.array([1, 2.5, 6.1]), 'order': np.array([1, 1, 2])}
               If the principle order number is 52, then these measured_lines represents 3 spectral features,
               with (pixel, diffraction order) coordinates of (1, 53), (2.5, 53), (6.1, 54).
               respectively. The wavelength solution will calibrate each fiber separately.
    :param reference_lines: list or ndarray. List of reference wavelengths in Angstroms
           for what features you expect are in measured_lines.
           Example:
               reference_lines = [5000, 5001, 5002, 5003.1, 6001.2]
    :param m0_range: tuple.
           The range of integers possible for the principle order number: (low, high). low is an inclusive bound
           and high is exclusive.
           The principle order number gives the real diffraction order index of the i=0 labelled diffraction order.
           See the xwavecal README or Brandt et al. (2019) for more.
    :param max_pixel: int. Maximum pixel coordinate on the detector. Used for finding the overlaps. E.g. 4096
    :param min_pixel: int. Minimum pixel coordinate on the detector. Used for finding the overlaps. E.g. 0
    :param wavelength_models: dict.
           Dictionary of models. See the xwavecal README, or wavelength_models below.
    :param overlap_settings: dict.
           Dictionary of settings for the overlap fitting algorithms.
           See the xwavecal README, or overlap_settings below.
    :param scale_settings: dict.
           Dictionary of settings for the global scale fitting algorithms.
           See the xwavecal README, or scale_settings below.
    :param stages_todo: list.
           list of xwavecal.wavelength.WavelengthSolution stages to run on the input list of measured lines.
           If building a new wavelength solution from scratch, then leave the default stages_todo.
    :return: measured_line_wavelengths: ndarray. if return_all==False.
             measured_line_wavelengths are the wavelengths of the measured_lines under the calibrated model.
             measured_line_wavelengths is an array with the same length as measured_lines['pixel']
              and measured_lines['order'].
    Notes
    -----
    See xwavecal.tests.test_wavelength.TestOnSyntheticData.test_performance for a use example of this function.
     the test can be run from an ipython instance and can provide insight as to how to use this wrapper function.
    """
    if stages_todo is None:
        stages_todo = [FitOverlaps, IdentifyPrincipleOrderNumber, SolveFromOverlaps,
                       FindGlobalScale, SolutionRefineInitial, SolutionRefineFinal]
    if wavelength_models is None:
        wavelength_models = {'initial_wavelength_model': {1: [0, 1, 2], 2: [0, 1, 2]},
                             'intermediate_wavelength_model': {0: [0, 1, 2], 1: [0, 1, 2], 2: [0, 1, 2]},
                             'final_wavelength_model': {0: [0, 1, 2, 3, 4, 5],
                                                        1: [0, 1, 2, 3, 4, 5],
                                                        2: [0, 1, 2, 3, 4, 5],
                                                        3: [0, 1, 2, 3, 4, 5],
                                                        4: [0]}}
    if overlap_settings is None:
        # settings relevant to fitting the overlaps.
        overlap_settings = {'min_num_overlaps': 5, 'overlap_linear_scale_range': (0.5, 2),
                            'flux_tol': 0.2, 'max_red_overlap': 1000, 'max_blue_overlap': 2000}
    if scale_settings is None:
        # settings relevant to finding the global scale.
        scale_settings = {'global_scale_range': (0.8, 1.2), 'approx_detector_range_angstroms': 5000,
                          'approx_num_orders': 67}
    # instantiate the context which every WavelengthStage will use.
    context = type('context', (), {**wavelength_models, **overlap_settings, **scale_settings,
                                   'm0_range': m0_range, 'main_spectrum_name':
                                   'spectrum', 'overlap_table_name': 'overlap'})

    # make dummy spectrum so that FitOverlaps will run.
    # TODO: the spectrum is only used in fit_overlaps to get the reference id's etc. It in principle is not
    #  needed at all. Fix this deeper in the code so that we don't have to make a fake spectrum here.
    pixel = np.arange(min_pixel, max_pixel + 1)
    orders = np.arange(np.min(measured_lines['order']), np.max(measured_lines['order'])).astype(int)
    spectrum_shape = (len(orders), len(pixel))
    spectrum = Table({'ref_id': orders, 'fiber': np.ones_like(orders),
                      'flux': np.zeros(spectrum_shape), 'pixel': pixel * np.ones(spectrum_shape)})
    # Initialize the WavelengthSolution
    wavelength_solution = WavelengthSolution(model=wavelength_models.get('initial_wavelength_model'),
                                             min_order=np.min(orders), max_order=np.max(orders),
                                             min_pixel=np.min(pixel), max_pixel=np.max(pixel),
                                             measured_lines=measured_lines,
                                             reference_lines=np.sort(reference_lines))
    # make a container (e.g. the Image object) for the spectrum and wavelength solution
    image = Image(header={'fiber_state': 'none&thar&none'},
                  wavelength_solution={1: wavelength_solution},
                  data_tables={context.main_spectrum_name: spectrum})
    # run the wavelength calibration stages
    for stage in stages_todo:
        image = stage(context).do_stage(image)
    measured_lines['wavelength'] = np.zeros_like(measured_lines['pixel'], dtype=float)
    wavelengths = None
    if image.wavelength_solution[1] is not None:
        wavelengths = image.wavelength_solution[1](measured_lines['pixel'], measured_lines['order'])
    measured_lines['wavelength'] = wavelengths
    return measured_lines['wavelength']




"""
main.py: Main driver script for the minimal pipeline included with this package.

Note: reduce_data() and run() are console entry points.
"""

import logging
import logging.config
from configparser import ConfigParser
from datetime import datetime
from ast import literal_eval
import pkg_resources
import os

from xwavecal.utils.runtime_utils import parse_args, get_data_paths, order_data, select_data_of_type, import_obj, safe_eval
from xwavecal.utils.fits_utils import Translator
from xwavecal.database import format_db_info, add_data_to_db

logging.config.fileConfig(pkg_resources.resource_filename('xwavecal', 'data/.logging.ini'))
logger = logging.getLogger(__name__)


class RuntimeContext(object):
    def __init__(self, dictionary):
        for attribute, value in dictionary.items():
            setattr(self, attribute, value)

    def __getattr__(self, attr: str):
        raise AttributeError('attribute {0} not found. Likely {0} is missing from the configuration file.'.format(attr))


def reduce_data(data_paths=None, args=None, config=None):
    if args is None:
        args = parse_args()
    if config is None:
        config = ConfigParser()
        config.read(args.config_file)
    if data_paths is None:
        data_paths = args.data_paths

    runtime_context, data_class, extension, header_keys, type_keys = organize_config(config)
    translator = Translator(header_keys, type_keys)
    DataClass = import_obj(data_class)
    data_paths = order_data(data_paths, DataClass, extension, header_keys, type_keys)

    for data_path in data_paths:
        logger.info('Reducing {path} assuming a data class of {data_class} and raw data in extension {extension}'
                    ''.format(path=data_path, data_class=data_class, extension=extension))

        data = DataClass.load(data_path, extension, translator)
        stages_todo = [import_obj(stage) for stage in literal_eval(config.get('stages', data.get_header_val('type')))]
        for stage in stages_todo:
            data = stage(runtime_context).do_stage(data)
            if isinstance(data, list) or isinstance(data, tuple):
                auxiliary_products = data[1:]
                data = data[0]
                for product in auxiliary_products:
                    write_out(product, runtime_context, args)

        write_out(data, runtime_context, args)


def run():
    # parse command line arguments and the configuration file.
    args = parse_args()
    config = ConfigParser()
    config.read(args.config_file)

    # get the data paths of the data to reduce.
    runtime_context, data_class, extension, header_keys, type_keys = organize_config(config)
    DataClass = import_obj(data_class)
    data_paths = select_data(args.input_dir, args.frame_type, literal_eval(config.get('data', 'files_contain')),
                             DataClass, extension, header_keys, type_keys)

    logger.info('Found {0} files of {1} type'.format(len(data_paths), args.frame_type))
    reduce_data(data_paths, args, config)


def write_out(data, runtime_context, args, use_db=True):
    data.filepath = make_output_path(args.output_dir, data, runtime_context.time_format)
    data.write(fpack=args.fpack)
    if use_db:
        logger.info('Adding file to processed image database at'
                    ' {path}'.format(path=runtime_context.database_path))
        db_info = format_db_info(data, runtime_context.time_format)
        add_data_to_db(runtime_context.database_path, db_info)


def organize_config(config):
    # pull information from the configuration file.
    runtime_context = RuntimeContext({key: safe_eval(item) for key, item in config.items('reduction')})
    data_class = config.get('data', 'data_class', fallback='xwavecal.images.Image')
    extension = config.getint('data', 'primary_data_extension')
    header_keys = literal_eval(config.get('data', 'header_keys'))
    type_keys = literal_eval(config.get('data', 'type_keys'))
    return runtime_context, data_class, extension, header_keys, type_keys


def select_data(input_dir, frame_type, files_contain, data_class, extension, header_keys, type_keys):
    data_paths = get_data_paths(input_dir, files_contain)
    data_paths = select_data_of_type(data_paths, data_class, extension, header_keys, type_keys, frame_type)
    return data_paths


def make_output_path(output_dir, data, time_fmt='%Y-%m-%dT%H:%M:%S.%f'):
    """
    :param data: Image
    :return: path: string
             The path to write the file to.
    """
    id = str(data.get_header_val('unique_id')).zfill(4)
    dayobs = datetime.strptime(data.get_header_val('observation_date'), time_fmt).strftime('%Y%m%d')
    filename = '{site}_{inst}_{dayobs}_{id}_{type}_{f0}{f1}{f2}.fits'.format(inst=data.get_header_val('instrument'),
                                                                             site=data.get_header_val('site_name'),
                                                                             dayobs=dayobs, id=id,
                                                                             type=data.get_header_val('type'),
                                                                             f0=data.fiber0_lit, f1=data.fiber1_lit,
                                                                             f2=data.fiber2_lit)
    filename = _sanitize(filename)
    return os.path.join(output_dir, filename)


def _sanitize(filename):
    filename = filename.replace('(', '_').replace(')', '_').replace(' ', '').replace(',', '_')
    return filename.replace('___', '_').replace('__', '_')



import mock
import numpy as np
from astropy.table import Table
from copy import deepcopy

from xwavecal.blaze import ApplyBlaze, BlazeMaker
from xwavecal.tests.utils import FakeContext, FakeImage
from xwavecal.utils.blaze_utils import normalize_orders


class TestApplyBlaze:
    CONTEXT = FakeContext()

    @mock.patch('os.path.exists', return_value=True)
    @mock.patch('xwavecal.blaze.ApplyBlaze.get_calibration_filename', return_value='fake')
    @mock.patch('xwavecal.images.Image.load')
    def test_apply_calibration(self, mock_load, mock_get_cal, mock_os):
        blaze = FakeImage()
        self.CONTEXT.min_blaze_sn = np.inf
        blaze.data = np.random.randint(1, 9, size=(10, 10))
        image = deepcopy(blaze)
        image.ivar = 4 * np.ones((10, 10))  # stderr = 1/2
        mock_load.return_value = blaze
        image = ApplyBlaze(self.CONTEXT).do_stage(image)
        assert np.allclose(image.data, 1)
        assert np.allclose(np.sqrt(image.ivar)**(-1), (1/2) / blaze.data)

    @mock.patch('xwavecal.blaze.ApplyBlaze.get_calibration_filename', return_value='fake')
    @mock.patch('xwavecal.images.Image.load')
    def test_apply_calibration_aborts_on_incorrect_shape(self, mock_load, mock_get_cal):
        blaze = FakeImage()
        blaze.data = np.random.randint(1, 9, size=(10, 11))
        image = FakeImage()
        image.data = np.ones((10, 10)).astype(float)
        mock_load.return_value = blaze
        image = ApplyBlaze(self.CONTEXT).do_stage(image)
        assert np.allclose(image.data, 1)


class TestBlazeMaker:
    CONTEXT = FakeContext()
    @mock.patch('xwavecal.blaze.normalize_orders')
    def test_do_stage(self, mock_normalize):
        mock_normalize.return_value = 2 * np.ones((3, 3), dtype=float)
        image = FakeImage()
        image.data = np.random.randint(1, 9, size=(3, 3))
        image, blaze = BlazeMaker(self.CONTEXT).do_stage(image)
        assert np.allclose(image.data / 2, blaze.data)
        assert image.get_header_val('type') == 'lampflat'
        assert blaze.get_header_val('type') == BlazeMaker(self.CONTEXT).calibration_type.lower()


def test_normalize_orders():
    data = np.ones((10, 100)).astype(float)
    data[4:7] *= 3
    data[1:3] *= 0.5
    trace = type('Trace', (), {'data': {'centers': [5 * np.ones((100))]}})
    normalization_factor = normalize_orders(data, trace, half_window=4, n=10)
    data /= normalization_factor
    assert np.allclose(data[4:7], 1)
    assert np.allclose(data[1:3], 0.5/3)
    # check that we set values which are not near the trace and which are not < minval to 0
    assert np.allclose([data[-1], data[0]], 0)
